import os
import pandas as pd
import time
from math import sqrt

mean = lambda data: float(sum(data) / len(data))
variance = lambda data: sum([x ** 2 for x in [i - mean(data) for i in data]]) / float(
    len(data)
)
std_dev = lambda data: sqrt(variance(data))


def save_csv(df, path, name, verbose=0, prompt=None):
    """saves a csv file from pandas DataFrame to the given path with the given name,
    if the entry is not a pandas DataFrame, it gets transformed to a pandas
    DataFrame before saving it

    Args:
        * df (pandas.DataFrame or array or numpy.array): A pandas.DataFrame or an array or a numpy.array
        * path (str): A string of the path where the file is going to be saved
        * name (str): A string of the name of the saved file with or without the .csv extention
        * verbose (int, optional): An integer of the verbosity of the function can be ``0`` or ``1``. Defaults to ``0``.
        * prompt (str, optional): A string of a custom prompt that is going to be displayed instead of the default
            generated prompt in case of verbosity set to ``1``. Defaults to ``None``.
    """
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)
    if not name.endswith(".csv"):
        name = name + ".csv"
    if not os.path.exists(path):
        os.makedirs(path)
    full_path = os.path.join(path, name)
    df.to_csv(full_path, index=False)
    if verbose > 0:
        if prompt is None:
            abs_path = os.path.join(os.getcwd(), path)
            print(
                f"\n\033[1mThe file {name} was saved in the path :\n{abs_path} \033[0m\n"
            )
        else:
            print(prompt)


def save_parquet(df, path, name, verbose=0, prompt=None):
    """saves a parquet file from pandas DataFrame to the given path with the given name,
    if the entry is not a pandas DataFrame, it gets transformed to a pandas
    DataFrame before saving it

    Args:
        * df (pandas.DataFrame or array or numpy.array): A pandas.DataFrame or an array or a numpy.array
        * path (str): A string of the path where the file is going to be saved
        * name (str): A string of the name of the saved file with or without the .parquet extention
        * verbose (int, optional): An integer of the verbosity of the function can be ``0`` or ``1``. Defaults to ``0``.
        * prompt (str, optional): A string of a custom prompt that is going to be displayed instead of the default
            generated prompt in case of verbosity set to ``1``. Defaults to ``None``.
    """
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)
    if not name.endswith(".parquet"):
        name = name + ".parquet"
    if not os.path.exists(path):
        os.makedirs(path)
    full_path = os.path.join(path, name)
    df.to_parquet(full_path, index=False)
    if verbose > 0:
        if prompt is None:
            abs_path = os.path.join(os.getcwd(), path)
            print(
                f"\n\033[1mThe file {name} was saved in the path :\n{abs_path} \033[0m\n"
            )
        else:
            print(prompt)


def read_csv(path, delimiter=" ", header=None, verbose=0, prompt=None):
    """A function to read a csv file and return a pandas dataframe.

    Args:
        * path (str): The data file name in the working directory or the data file path with the file name.
        * delimiter (str, optional): A string for the type of separation used in the csv file. Defaults to ``" "``.
        * header (int, optional): The specification of the technique used to define the columns names: ``None`` in case of no columns names in the file, ``0`` in case that first row is the header. Defaults to ``None``.
        * verbose (int, optional): An integer of the verbosity of the function can be ``0`` or ``1``. Defaults to ``0``.
        * prompt (str, optional): A message in the case of verbose is ``1``, if not specified another default message will be displayed. Defaults to ``None``.

    Returns:
        * pandas.DataFrame: A pandas dataframe of the read file.
    """
    df_csv = pd.read_table(path, header=header, delimiter=delimiter)
    if verbose > 0:
        if prompt is None:
            print(f"\n\033[1mLoading dataframe from csv file in:\n{path} \033[0m\n")
        else:
            print(prompt)
    return df_csv


def read_parquet(path, verbose=0, prompt=None):
    """A function to read a parquet file and return a pandas dataframe.

    Args:
        * path (str): The data file name in the working directory or the data file path with the file name.
        * delimiter (str, optional): A string for the type of separation used in the parquet file. Defaults to ``" "``.
        * verbose (int, optional): An integer of the verbosity of the function can be ``0`` or ``1``. Defaults to ``0``.
        * prompt (str, optional): A message in the case of verbose is ``1``, if not specified another default message will be displayed. Defaults to ``None``.

    Returns:
        * pandas.DataFrame: A pandas dataframe of the read file.
    """
    df_parquet = pd.read_parquet(path)
    if verbose > 0:
        if prompt is None:
            print(f"\n\033[1mLoading dataframe from parquet file in:\n{path} \033[0m\n")
        else:
            print(prompt)
    return df_parquet


def dict_transpose(dictionary):
    """A function that transposes a dictionary,
    it simply uses the first key and it's values as the new keys and then maps the rest of the keys and their values to the newly created keys in the same order of apperance.

    Args:
        * dictionary (dict): A python dictionary

    Returns:
        * dict: A transposed python dictionary

    Example:
        >>> d = {
            "classifier": ["SVM","LR","MLP"],
            "scaler": ["Standard", "Standard", "Standard"],
            "exec time": ["75.88(s)", "4.78(s)", "94.89(s)"],
            "accuracy": ["78.5%","53.6%","88.6%"],
            "F1": ["78.6%","53.0%","88.6%"],
            }
        >>> d_transposed = dict_transpose(d)
        >>> d_transposed
            {
            "classifier": ["scaler","exec time","accuracy","F1"],
            "SVM": ["Standard","75.88(s)","78.5%","78.6%"],
            "LR": ["Standard","4.78(s)","53.6%","53.0%"],
            "MLP": ["Standard","94.89(s)","88.6%","88.6%"],
            }
    """
    keys_list = list(dictionary)
    values_list = list(dictionary.values())
    new_dict = {keys_list[0]: keys_list[1:]}
    new_keys = values_list[0]
    for key in new_keys:
        new_dict[key] = []
    for values in values_list[1:]:
        for key, v in zip(new_keys, values):
            new_dict[key].append(v)
    return new_dict


class FileScraper:
    """A class for file scraping from a specific directory. includes ``path_list`` attribute representing the paths of the searched files in a list of strings format.
    ``found_files_count`` attribute represnting the number of files found.
    `all_files` attribute represnting the paths of all existing files in a list of strings format available in the given path directory.
    ``all_files_count`` attribute representing the number of all existing files available in the given path directory.
    `time` attribute representing the execution time of the search.
    A `__call__ ` method for calling an instance of the class to return the `path_list` attribute.


    Args:
        * path (str): The path to the search directory.
        * search_list (list): List of strings representing the searched files.
        * verbose (int, optional): An integer of the verbosity of the function can be ``0`` or ``1``. Defaults to ``0``.
    """

    def __init__(self, path, search_list, verbose=0) -> None:
        """The constructor of the FileScraper class.

        Args:
            * path (str): The path to the search directory.
            * search_list (list): List of strings representing the searched files.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1`` or ``2`` or ``3``. Defaults to ``0``.
        """
        self.verbose = verbose
        start_time = time.perf_counter()
        self.parent_path = path
        self.search_list = set(search_list)
        self.searched_list = search_list.copy()
        self.file_scraping(self.parent_path, self.search_list)
        self.found_files_count = len(self.path_list)
        self.all_files_count = len(self.all_files)
        end_time = time.perf_counter()
        self.time = f"{end_time-start_time} (s)"
        if self.verbose == 1:
            print(f"Finished searching in {self.time}")
            print(
                f"{self.found_files_count} matching files found from total of {self.all_files_count} existant files"
            )
        if len(self.searched_list):
            print(f"These elements are not found: {self.searched_list}\n")
        else:
            if self.verbose == 1:
                print("*******All searched elements were found successfully*******\n")

    def __call__(self):
        """A method for the class instance call

        Returns:
            * list: list of strings representing the paths of the searched files.
        """
        return self.path_list

    def file_scraping(self, path, search_list):
        """A class method that searches

        Args:
            * path (str): The path to the search directory.

        Returns:
            list: list of strings representing the paths of the searched files.
        """
        files = []
        for text in os.listdir(path):
            if os.path.isdir(os.path.join(path, text)):
                if self.verbose == 3:
                    print(f"Changing directory to subdirectory: '{text}'")
                dir = os.path.join(path, text)
                localfiles = self.file_scraping(dir, search_list)
                if self.verbose == 3:
                    if localfiles != []:
                        print(f"'{text}' files :{localfiles}")
                if localfiles == []:
                    localfiles, dir = files, path
                    if self.verbose == 3:
                        if localfiles == []:
                            print(f"No files found in the '{text}' directory")
                        else:
                            print(f"Changing directory to main directory:")
                            print(f"'{dir}'\nfiles :{localfiles}")
                for file in localfiles:
                    file_path = os.path.join(dir, file)
                    try:
                        if file not in self.all_files:
                            self.all_files.append(file_path)
                    except AttributeError:
                        self.all_files = [file_path]
                    for picked_file in search_list:
                        if file == picked_file:
                            if picked_file in self.searched_list:
                                self.searched_list.remove(picked_file)
                            try:
                                if file_path not in self.path_list:
                                    self.path_list.append(file_path)
                                    if self.verbose == 2:
                                        print(f"file '{file}' found in: \n{dir}")
                                        print(
                                            "**File path added to the path_list successfully**\n"
                                        )
                            except AttributeError:
                                self.path_list = [file_path]
                                if self.verbose == 2:
                                    print(f"file '{file}' found in: \n{dir}")
                                    print(
                                        "**Initalization of the path_list with the file path is successful**\n"
                                    )

            elif os.path.isfile(os.path.join(path, text)):
                files.append(text)
        return files


def pattern_search(pattern, local_set, error_message, global_set=None):
    """A function to collect phrases starting from a string pattern.

    Args:
        * pattern (str): A sylable or a part of a word that we want to find.
        * local_set (set): a set or list of complete and correct words that the pattern could be in.
        * error_message (str): a message to print with the ValueError raised in case pattern is not found.
        * global_set (set): a set or a list of phrases that we want to find the word that match the pattern in. If set to `None`, the search is condacted only in the local_set. Defaults to `None`.

    Returns:
        * set: a set of the phrases that contain the word matching the entry pattern.
    """
    matching_result_set = set()
    for word in local_set:
        if pattern in word:
            if global_set:
                for element in global_set:
                    if word in element:
                        matching_result_set.add(element)
            else:
                matching_result_set.add(word)
    if len(matching_result_set):
        return matching_result_set
    else:
        raise ValueError(error_message)


if __name__ == "__main__":
    DATA_DIR_PATH = "/run/media/najem/34b207a8-0f0c-4398-bba2-f31339727706/home/stock/The_stock/dev & datasets/PhD/datasets/UAH-DRIVESET-v1"
    # 'RAW_ACCELEROMETERS.txt' "RAW_GPS.txt",
    file_list = FileScraper(
        DATA_DIR_PATH,
        search_list=["RAW_ACCELEROMETERS.txt", "RAW_GPS.txt", "try"],
        verbose=0,
    )
    print(file_list())



from keras.layers import (
    Input,
    Conv2D,
    BatchNormalization,
    MaxPooling2D,
    Dropout,
    Flatten,
    Dense,
    Add,
    Activation,
)
import time
import platform
from matplotlib import pyplot
from keras.models import Model, load_model, save_model
from keras.callbacks import ModelCheckpoint, LearningRateScheduler
from maaml.machine_learning import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    cohen_kappa_score,
    roc_auc_score,
    train_test_split,
    ShuffleSplit,
    np,
)
from maaml.utils import save_csv


class DeepRCNModel:
    """A class for the DeepRCNModel model, that allow you to access the layers as attributes by the ordinal numbers names from first_layer .. to the eighteenth_layer, except for the attributes of the resnet_block that has it's unique method and the input_layer and output_layer attributes.
    Includes a static method attribute that can be independently accessed to create the built-in resnet block function, and a `__call__ ` method for calling an instance of the class to return the `fmodel` attribute.

    Args:
        * input_shape (tuple, optional): The input shape of the model. Defaults to `(20, 1, 1)`.
        * class_nb (int, optional): The number of targets in the outputs . Defaults to `3`.
    """

    def __init__(self, input_shape=(20, 1, 1), class_nb=3):
        """The constructor of the DeepRCNModel Class

        Args:
            * input_shape (tuple, optional): The input shape of the model. Defaults to `(20, 1, 1)`.
            * class_nb (int, optional): The number of targets in the outputs . Defaults to `3`.
        """
        self.input_shape = input_shape
        self.input_layer = Input(shape=self.input_shape, name="input")
        self.first_layer = Conv2D(
            60, (4, 4), activation="relu", padding="same", name="conv1_60_4x4"
        )(self.input_layer)
        self.second_layer = BatchNormalization(name="batch_norm1")(self.first_layer)
        self.third_layer = MaxPooling2D((2, 2), padding="same", name="Max_pool1_2x2")(
            self.second_layer
        )
        self.fourth_layer = Dropout(0.2, name="dropout1_0.2")(self.third_layer)
        self.resnet_block = self.resnet_block_creating(self.fourth_layer, 60, 4)
        self.fifth_layer = Conv2D(
            30, (4, 4), activation="relu", padding="same", name="conv2_30_4x4"
        )(self.resnet_block)
        self.sixth_layer = BatchNormalization(name="batch_norm2")(self.fifth_layer)
        self.seventh_layer = MaxPooling2D((2, 2), padding="same", name="Max_pool2_2x2")(
            self.sixth_layer
        )
        self.eighth_layer = Dropout(0.2, name="dropout2_0.2")(self.seventh_layer)
        self.ninth_layer = Conv2D(
            15, (4, 4), activation="relu", padding="same", name="conv3_15_4x4"
        )(self.eighth_layer)
        self.tenth_layer = BatchNormalization(name="batch_norm3")(self.ninth_layer)
        self.eleventh_layer = MaxPooling2D(
            (2, 2), padding="same", name="Max_pool3_2x2"
        )(self.tenth_layer)
        self.twelfth_layer = Dropout(0.2, name="dropout3_0.2")(self.eleventh_layer)
        self.thirdteenth_layer = Conv2D(
            8, (4, 4), activation="relu", padding="same", name="conv4_8_4x4"
        )(self.twelfth_layer)
        self.fourteenth_layer = BatchNormalization(name="batch_norm4")(
            self.thirdteenth_layer
        )
        self.fifteenth_layer = MaxPooling2D(
            (1, 1), padding="same", name="Max_pool4_1x1"
        )(self.fourteenth_layer)
        self.sixteenth_layer = Flatten(name="Flatten_layer")(self.fifteenth_layer)
        self.seventeenth_layer = Dense(
            units=224, input_dim=448, activation="relu", name="dense1_448x224"
        )(self.sixteenth_layer)
        self.eighteenth_layer = BatchNormalization(name="batch_norm5")(
            self.seventeenth_layer
        )
        self.output_layer = Dense(
            units=class_nb, activation="softmax", name="dense2_224x10"
        )(self.eighteenth_layer)
        self.model = Model(self.input_layer, self.output_layer, name="DeepRCNModel")

    def __call__(self):
        """A method for the class instance call

        Returns:
            * keras class: The keras model.
        """
        return self.model

    @staticmethod
    def resnet_block_creating(input_layer, filters, conv_size):
        """A static method to create an internal resnet block that links with the global model layers.

        Args:
            * input_layer (tensorflow or keras class): A model layer that the resnet block is going to link to.
            * filters (int): The number of filters in the Conv2D internal layers.
            * conv_size (int): The convolution size of the Conv2D internal layers.

        Returns:
            * A tensorflow or keras class: A layer that links to a block of layers.
        """
        x = Conv2D(
            filters, conv_size, activation="relu", padding="same", name="resnet_block1"
        )(input_layer)
        x = BatchNormalization(name="resnet_block2")(x)
        x = Conv2D(
            filters, conv_size, activation=None, padding="same", name="resnet_block3"
        )(x)
        x = BatchNormalization(name="resnet_block4")(x)
        x = Dropout(0.2)(x)
        x = Add(name="resnet")([x, input_layer])
        x = Activation("relu", name="resnet_activation")(x)
        return x

    def show(self):
        """A method to display the model's summary."""
        self.model.summary()


class Evaluator:
    """A class to evaluate a tensorflow or a keras deep learning model and saves the trained model and the results in the working directory.

    Args:
        * model (tensorflow or a keras class): A model to evaluate
        * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features and a classfication target in one column and the target in one hot encoded format. Defaults to `None`.
        * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
        * target_column (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
        * target (pandas.DataFrame or array or numpy.array, optional): The classfication target in one hot encoded format in the case that the dataset parameter is not provided. Defaults to `None`.
        * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
        * model_name (str, optional): The name of the model as string that is going to be displayed in the cross validation results. Defaults to `None`.
        * input_shape (tuple): The input shape of the model as a tuple. Defaults to `None`.
        * preprocessing_alias (str, optional): The name for the applied dataset preprocessing as a string that is going to be displayed in the cross validation results. Defaults to `None`.
        * cross_eval (bool, optional): A full cross validation evaluation in the case of `True` or a training session evaluation in the case of `False`. Defaults to `True`.
        * save_tag (str, optional): The tag given to the created directory and the saved evaluation result, important in case of mutiple evaluation in the same directory to not overwite exiting results. Defaults to `None`.
        * nb_splits (int, optional): The number of splits in case of cross validation. Defaults to `5`.
        * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
        * callbacks (str, optional): The training model callback specification that saves the training model if not specified as `"best model"`, the trained model will not be saved. Defaults to `"best model"`.
        * learning_rate_scheduler (str, optional): If defined as `"scheduler"` uses the internal `learning_rate_sheduling` method to define and change the learning rate. Defaults to `"scheduler"`.
        * opt (str or a tensorflow or keras class, optional): The optimizer used in the evaluation. Defaults to `"adam"`. exemple: if defined as `"adam"` uses the standard Adam optimizer class available in keras with it's standard learning rate of `0.001`.
        * loss (str or a tensorflow or keras class, optional): The loss function used in the evaluation. Defaults to `"categorical_crossentropy"`. exemple if defined as `"categorical_crossentropy"` uses the standard categorical crossentropy loss class available in keras.
        * metrics (list, optional): The metric used in the training can be a list of string or list of tensorflow or keras classes, this parameter is only used in the training and does not reflect the evaluation metrics in the results that includes various other metrics for the results. Defaults to `["accuracy"]`.
        * epochs (int, optional): The number of epochs for the training. Defaults to `600`.
        * batch_size (int, optional): The size of the batch in the training. Defaults to `60`.
        * verbose (int, optional): An integer of the verbosity of the evaluation can be ``0`` or ``1``. Defaults to ``0``.
    """

    def __init__(
        self,
        model,
        dataset=None,
        features=None,
        target_column=None,
        target=None,
        target_name="target",
        model_name: str = None,
        input_shape=None,
        preprocessing_alias=None,
        cross_eval=True,
        save_tag=None,
        nb_splits=5,
        test_size=0.3,
        callbacks="best model",
        learning_rate_scheduler="scheduler",
        opt="adam",
        loss="categorical_crossentropy",
        metrics=["accuracy"],
        epochs=600,
        batch_size=60,
        verbose=1,
    ):
        """The constructor of the Evaluator class.

        Args:
            * model (tensorflow or a keras class): A model to evaluate
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features and a classfication target in one column and the target in one hot encoded format. Defaults to `None`.
            * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
            * target_column (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
            * target (pandas.DataFrame or array or numpy.array, optional): The classfication target in one hot encoded format in the case that the dataset parameter is not provided. Defaults to `None`.
            * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
            * model_name (str, optional): The name of the model as string that is going to be displayed in the cross validation results. Defaults to `None`.
            * input_shape (tuple): The input shape of the model as a tuple. Defaults to `None`.
            * preprocessing_alias (str, optional): The name for the applied dataset preprocessing as a string that is going to be displayed in the cross validation results. Defaults to `None`.
            * cross_eval (bool, optional): A full cross validation evaluation in the case of `True` or a training session evaluation in the case of `False`. Defaults to `True`.
            * save_tag (str, optional): The tag given to the created directory and the saved evaluation result, important in case of mutiple evaluation in the same directory to not overwite exiting results. Defaults to `None`.
            * nb_splits (int, optional): The number of splits in case of cross validation. Defaults to `5`.
            * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
            * callbacks (str, optional): The training model callback specification that saves the training model if not specified as `"best model"`, the trained model will not be saved. Defaults to `"best model"`.
            * learning_rate_scheduler (str, optional): If defined as `"scheduler"` uses the internal `learning_rate_sheduling` method to define and change the learning rate. Defaults to `"scheduler"`.
            * opt (str or a tensorflow or keras class, optional): The optimizer used in the evaluation. Defaults to `"adam"`. exemple: if defined as `"adam"` uses the standard Adam optimizer class available in keras with it's standard learning rate of `0.001`.
            * loss (str or a tensorflow or keras class, optional): The loss function used in the evaluation. Defaults to `"categorical_crossentropy"`. exemple if defined as `"categorical_crossentropy"` uses the standard categorical crossentropy loss class available in keras.
            * metrics (list, optional): The metric used in the training can be a list of string or list of tensorflow or keras classes, this parameter is only used in the training and does not reflect the evaluation metrics in the results that includes various other metrics for the results. Defaults to `["accuracy"]`.
            * epochs (int, optional): The number of epochs for the training. Defaults to `600`.
            * batch_size (int, optional): The size of the batch in the training. Defaults to `60`.
            * verbose (int, optional): An integer of the verbosity of the evaluation can be ``0`` or ``1``. Defaults to ``0``.
        """
        self.save_tag = "" if save_tag is None or save_tag == "" else f"_{save_tag}"
        if preprocessing_alias is not None:
            self.save_tag = f"_{preprocessing_alias}" + self.save_tag
        else:
            preprocessing_alias = ""
        self.model_name = model_name if model_name is not None else model.name
        target_name = [target_name]
        self.target_list = []
        if dataset is not None:
            for column_name in dataset.columns:
                for keyname in target_name:
                    if keyname in column_name:
                        self.target_list.append(column_name)
        if cross_eval is True:
            self.cross_evaluation = self.cross_validating(
                model=model,
                dataset=dataset,
                features=features,
                target_column=target_column,
                target=target,
                target_names=self.target_list,
                model_name=self.model_name,
                preprocessing_alias=preprocessing_alias,
                input_shape=input_shape,
                save_tag=self.save_tag,
                callbacks=callbacks,
                learning_rate_scheduler=learning_rate_scheduler,
                nb_splits=nb_splits,
                test_size=test_size,
                opt=opt,
                loss=loss,
                metrics=metrics,
                epochs=epochs,
                batch_size=batch_size,
                verbose=verbose,
            )
        elif cross_eval is not True:
            (
                self.trained_model,
                self.best_model,
                self.training_history,
                self.evaluation,
            ) = self.model_training(
                model=model,
                dataset=dataset,
                features=features,
                target_column=target_column,
                target=target,
                target_names=self.target_list,
                model_name=self.model_name,
                input_shape=input_shape,
                save_tag=self.save_tag,
                callbacks=callbacks,
                learning_rate_scheduler=learning_rate_scheduler,
                test_size=test_size,
                opt=opt,
                loss=loss,
                metrics=metrics,
                epochs=epochs,
                batch_size=batch_size,
                verbose=verbose,
            )

    def model_training(
        self,
        model,
        dataset,
        features,
        target_column,
        target,
        target_names,
        input_shape,
        save_tag="",
        callbacks="best model",
        learning_rate_scheduler="scheduler",
        test_size=0.3,
        opt="adam",
        loss="categorical_crossentropy",
        metrics=["accuracy"],
        epochs=600,
        batch_size=60,
        verbose=1,
    ):
        """A method for training a tensorflow or a keras deep learning model and saves the trained model and the results in the working directory.

        Args:
            * model (tensorflow or a keras class): A model to evaluate
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features and a classfication target in one column and the target in one hot encoded format. Defaults to `None`.
            * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
            * target_column (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
            * target (pandas.DataFrame or array or numpy.array, optional): The classfication target in one hot encoded format in the case that the dataset parameter is not provided. Defaults to `None`.
            * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
            * model_name (str, optional): The name of the model as string that is going to be displayed in the training results. Defaults to `None`.
            * input_shape (tuple): The input shape of the model as a tuple. Defaults to `None`.
            * save_tag (str, optional): The tag given to the created directory and the saved evaluation result, important in case of mutiple evaluation in the same directory to not overwite exiting results. Defaults to `None`.
            * callbacks (str, optional): The training model callback specification that saves the training model if not specified as `"best model"`, the trained model will not be saved. Defaults to `"best model"`.
            * learning_rate_scheduler (str, optional): If defined as `"scheduler"` uses the internal `learning_rate_sheduling` method to define and change the learning rate. Defaults to `"scheduler"`.
            * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
            * opt (str or a tensorflow or keras class, optional): The optimizer used in the evaluation. Defaults to `"adam"`. exemple: if defined as `"adam"` uses the standard Adam optimizer class available in keras with it's standard learning rate of `0.001`.
            * loss (str or a tensorflow or keras class, optional): The loss function used in the evaluation. Defaults to `"categorical_crossentropy"`. exemple if defined as `"categorical_crossentropy"` uses the standard categorical crossentropy loss class available in keras.
            * metrics (list, optional): The metric used in the training can be a list of string or list of tensorflow or keras classes, this parameter is only used in the training and does not reflect the evaluation metrics in the results that includes various other metrics for the results. Defaults to `["accuracy"]`.
            * epochs (int, optional): The number of epochs for the training. Defaults to `600`.
            * batch_size (int, optional): The size of the batch in the training. Defaults to `60`.
            * verbose (int, optional): An integer of the verbosity of the evaluation can be ``0`` or ``1``. Defaults to ``0``.

        Returns:
            * tuple: (`model`, `best_model`, `training_history`, `scores`)
                * tensorflow or keras class: `model` the model after training as the first value.
                * tensorflow or keras class: `best_model` the best performing the model as the second value.
                * dict: `training_history` the training history as the third value.
                * list: `scores` the result model evaluation scores as the fourth value.
        """
        start_time = time.perf_counter()
        if dataset is not None:
            X = dataset.drop(target_names, axis=1)
            Y = dataset[target_names[0]]
            Y_ohe = dataset[target_names[1:]]
        elif features is not None and target is not None:
            X, Y, Y_ohe = features, target_column, target
        X = np.reshape(X.values, (len(X), *input_shape))
        if any(c.isdigit() for c in platform.node()) == True:
            PATH = "/content/drive/MyDrive/"
        else:
            PATH = f"DL_EVALUATION{save_tag}/"
        if callbacks == "best model":
            filepath = PATH + "best_model.h5"
            mc = ModelCheckpoint(
                filepath, monitor="val_accuracy", save_best_only=True, verbose=1
            )
            cb = [mc]
            if learning_rate_scheduler == "scheduler":
                lrs = LearningRateScheduler(self.learning_rate_sheduling)
                cb.append(lrs)
        elif callbacks == None:
            cb = None
        model.compile(loss=loss, optimizer=opt, metrics=metrics)
        X_train, X_test, Y_ohe_train, Y_ohe_test, _, Y_test = train_test_split(
            X, Y_ohe, Y, test_size=test_size, random_state=10
        )
        history = model.fit(
            X_train,
            Y_ohe_train,
            validation_data=(X_test, Y_ohe_test),
            callbacks=cb,
            epochs=epochs,
            batch_size=batch_size,
            verbose=verbose,
        )
        end_time = time.perf_counter()
        training_history = history.history
        save_csv(training_history, PATH, "training_history")
        if callbacks == "best model":
            try:
                best_model = load_model(filepath)
            except Exception:
                try:
                    best_model = load_model("best_model.h5")
                except Exception:
                    print(
                        "Can't load the saved best model,revert to best_model = model"
                    )
                    best_model = model
        pred = best_model.predict(X_test, batch_size=batch_size, verbose=1)
        predictions = np.argmax(pred, axis=1)
        train_eval = best_model.evaluate(X_train, Y_ohe_train, verbose=verbose)
        train_score = train_eval[1]
        exec_time = f"{(end_time - start_time): .2f} (s)"
        acc_score = accuracy_score(Y_test.values, predictions, normalize=True)
        pres_score = precision_score(Y_test.values, predictions, average="macro")
        rec_score = recall_score(Y_test.values, predictions, average="macro")
        f1 = f1_score(Y_test.values, predictions, average="macro")
        cokap_score = cohen_kappa_score(Y_test.values, predictions)
        roc_auc = roc_auc_score(Y_ohe_test.values, pred)
        scores = [
            f"execution time: {exec_time}",
            f"train accuracy: {train_score: .4%} ",
            f"accuracy: {acc_score: .4%}",
            f"precision: {pres_score: .4%}",
            f"recall: {rec_score: .4%}",
            f"F1 score: {f1: .4%}",
            f"cohen kappa: {cokap_score: .4%}",
            f"roc_auc_score: {roc_auc: .4%}",
        ]
        if verbose == 1:
            print(scores)
        return model, best_model, training_history, scores

    def cross_validating(
        self,
        model,
        dataset,
        features,
        target_column,
        target,
        target_names,
        model_name,
        input_shape,
        preprocessing_alias="",
        save_tag="",
        callbacks="best model",
        learning_rate_scheduler="scheduler",
        nb_splits=5,
        test_size=0.3,
        opt="adam",
        loss="categorical_crossentropy",
        metrics=["accuracy"],
        epochs=600,
        batch_size=60,
        verbose=0,
    ):
        """A method for cross evaluating and training a tensorflow or a keras deep learning model and saves the trained models and the results in the working directory.

        Args:
            * model (tensorflow or a keras class): A model to evaluate
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features and a classfication target in one column and the target in one hot encoded format. Defaults to `None`.
            * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
            * target_column (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
            * target (pandas.DataFrame or array or numpy.array, optional): The classfication target in one hot encoded format in the case that the dataset parameter is not provided. Defaults to `None`.
            * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
            * model_name (str, optional): The name of the model as string that is going to be displayed in the cross validation results. Defaults to `None`.
            * input_shape (tuple): The input shape of the model as a tuple. Defaults to `None`.
            * preprocessing_alias (str, optional): The name for the applied dataset preprocessing as a string that is going to be displayed in the cross validation results. Defaults to `None`.
            * save_tag (str, optional): The tag given to the created directory and the saved evaluation result, important in case of mutiple evaluation in the same directory to not overwite exiting results. Defaults to `None`.
            * callbacks (str, optional): The training model callback specification that saves the training model if not specified as `"best model"`, the trained model will not be saved. Defaults to `"best model"`.
            * learning_rate_scheduler (str, optional): If defined as `"scheduler"` uses the internal `learning_rate_sheduling` method to define and change the learning rate. Defaults to `"scheduler"`.
            * nb_splits (int, optional): The number of splits in case of cross validation. Defaults to `5`.
            * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
            * opt (str or a tensorflow or keras class, optional): The optimizer used in the evaluation. Defaults to `"adam"`. exemple: if defined as `"adam"` uses the standard Adam optimizer class available in keras with it's standard learning rate of `0.001`.
            * loss (str or a tensorflow or keras class, optional): The loss function used in the evaluation. Defaults to `"categorical_crossentropy"`. exemple if defined as `"categorical_crossentropy"` uses the standard categorical crossentropy loss class available in keras.
            * metrics (list, optional): The metric used in the training can be a list of string or list of tensorflow or keras classes, this parameter is only used in the training and does not reflect the evaluation metrics in the results that includes various other metrics for the results. Defaults to `["accuracy"]`.
            * epochs (int, optional): The number of epochs for the training. Defaults to `600`.
            * batch_size (int, optional): The size of the batch in the training. Defaults to `60`.
            * verbose (int, optional): An integer of the verbosity of the evaluation can be ``0`` or ``1``. Defaults to ``0``.

        Returns:
            * list: `cv_scores` the result cross validation evaluation scores.
        """
        start_time = time.perf_counter()
        if dataset is not None:
            X = dataset.drop(target_names, axis=1)
            Y = dataset[target_names[0]]
            Y_ohe = dataset[target_names[1:]]
        elif features is not None and target is not None:
            X, Y, Y_ohe = features, target_column, target
        X = np.reshape(X.values, (len(X), *input_shape))
        if any(c.isdigit() for c in platform.node()) == True:
            PATH = "/content/drive/MyDrive/"
        else:
            PATH = f"DL_EVALUATION{save_tag}/"
        model_path = PATH + "base_model.h5"
        save_model(model, model_path)
        cv = ShuffleSplit(n_splits=nb_splits, test_size=test_size, random_state=10)
        (
            exec_time,
            train_acc_scores,
            acc_scores,
            pres_scores,
            rec_scores,
            f1,
            cokap_scores,
            roc_auc_scores,
        ) = ([], [], [], [], [], [], [], [])
        cv_scores = {}
        for train, test in cv.split(X, Y, Y_ohe):
            prompt_message = f"\033[1m\n*******begin cross validation in fold number:{len(train_acc_scores) + 1}*******\033[0m"
            print(prompt_message)
            if callbacks == "best model":
                filepath = PATH + f"cv_best_model{len(train_acc_scores) + 1}.h5"
                mc = ModelCheckpoint(
                    filepath, monitor="val_accuracy", save_best_only=True, verbose=1
                )
                cb = [mc]
                if learning_rate_scheduler == "scheduler":
                    lrs = LearningRateScheduler(self.learning_rate_sheduling)
                    cb.append(lrs)
            elif callbacks == None:
                cb = None
            cv_model = load_model(model_path, compile=False)
            cv_model.compile(loss=loss, optimizer=opt, metrics=metrics)
            history = cv_model.fit(
                X[train],
                Y_ohe.loc[train],
                validation_data=(X[test], Y_ohe.loc[test]),
                callbacks=cb,
                epochs=epochs,
                batch_size=batch_size,
                verbose=verbose,
            )
            end_time = time.perf_counter()
            training_history = history.history
            save_csv(
                training_history,
                PATH,
                f"training_history_cv{len(train_acc_scores) + 1}",
            )
            if callbacks == "best model":
                try:
                    best_model = load_model(filepath)
                except Exception:
                    print(
                        "Can't load the saved best model,revert to best_model = cv_model"
                    )
                    best_model = cv_model
            elif callbacks == None:
                best_model = cv_model
                saved_model_path = PATH + f"cv_model{len(train_acc_scores) + 1}.h5"
                save_model(best_model, saved_model_path)
            pred = best_model.predict(X[test], batch_size=batch_size, verbose=0)
            predictions = np.argmax(pred, axis=1)
            exec_time.append((end_time - start_time))
            train_acc_scores.append(
                best_model.evaluate(X[train], Y_ohe.loc[train], verbose=verbose)[1]
            )
            acc_scores.append(
                accuracy_score(Y[test].values, predictions, normalize=True)
            )
            pres_scores.append(
                precision_score(Y[test].values, predictions, average="macro")
            )
            rec_scores.append(
                recall_score(Y[test].values, predictions, average="macro")
            )
            f1.append(f1_score(Y[test].values, predictions, average="macro"))
            cokap_scores.append(cohen_kappa_score(Y[test].values, predictions))
            roc_auc_scores.append(roc_auc_score(Y_ohe.loc[test].values, pred))
        cv_scores["metrics"] = [
            "preprocessing",
            "execution time",
            "training accuracy",
            "accuracy",
            "precision",
            "recall",
            "F1",
            "cohen_kappa",
            "roc_auc",
        ]
        cv_scores[model_name] = [
            preprocessing_alias,
            f"{np.mean(exec_time): .2f} (s)",
            f"{np.mean(train_acc_scores):.4%} (+/- {np.std(train_acc_scores):.4%})",
            f"{np.mean(acc_scores):.4%} (+/- {np.std(acc_scores):.4%})",
            f"{np.mean(pres_scores):.4%} (+/- {np.std(pres_scores):.4%})",
            f"{np.mean(rec_scores):.4%} (+/- {np.std(rec_scores):.4%})",
            f"{np.mean(f1):.4%} (+/- {np.std(f1):.4%})",
            f"{np.mean(cokap_scores):.4%} (+/- {np.std(cokap_scores):.4%})",
            f"{np.mean(roc_auc_scores):.4%} (+/- {np.std(roc_auc_scores):.4%})",
        ]
        save_csv(cv_scores, PATH, f"cross_validation_{preprocessing_alias}{save_tag}")
        return cv_scores

    @staticmethod
    def learning_rate_sheduling(
        epoch, lrate, initial_lrate=0.001, second_lrate=0.0001, scheduler_threshold=480
    ) -> float:
        """A static method to customize a learning rate sheduling.

        Args:
            * epoch (int): The current epoch count in the training process.
            * lrate (float):The current learning rate in the training process.
            * initial_lrate (float, optional): The chosen initial learning rate to be set for the training. Defaults to `0.001`.
            * second_lrate (float, optional): The chosen learning rate to change the training to when the threshold is reached. Defaults to `0.0001`.
            * scheduler_threshold (int, optional): The epoch value that defines when the change in the learning rate will occur. Defaults to `480`.

        Returns:
            * float: The actual learning rate that is going to be used in the training process.
        """
        lrate = initial_lrate
        if epoch > scheduler_threshold:
            lrate = second_lrate
            print("Change in the learning rate, the new learning rate is:", lrate)
        return lrate

    @staticmethod
    def training_history_ploting(
        training_history, save=False, metric="accuracy", style="default"
    ):
        """A static method for ploting the history of a training session

        Args:
            * training_history (pandas.DataFrame): A training history of a tensorflow or keras class model training session in the form of pandas dataframe.
            * save (bool, optional): saves the plot in the current working directory in the case of `True`, and does not save the plot in the case of `False`. Defaults to `False`.
            * metric (str, optional): the metric that we want to display available in the training history dataframe columns such as `"accuracy"` or `"loss"`. Defaults to `"accuracy"`.
            * style (str, optional): The style of the displayed plot, can be a name from the stanadard matplotlib library styles such as `"seaborn-poster"` or an integer from `"0"` to `"23"` passed as a string . Defaults to "default".
        """
        styles = [
            "seaborn-poster",
            "seaborn-bright",
            "Solarize_Light2",
            "seaborn-whitegrid",
            "_classic_test_patch",
            "seaborn-white",
            "fivethirtyeight",
            "seaborn-deep",
            "seaborn",
            "seaborn-dark-palette",
            "seaborn-paper",
            "seaborn-darkgrid",
            "seaborn-notebook",
            "grayscale",
            "seaborn-muted",
            "seaborn-dark",
            "seaborn-talk",
            "ggplot",
            "bmh",
            "dark_background",
            "fast",
            "seaborn-ticks",
            "seaborn-colorblind",
            "classic",
        ]
        if type(style) is int:
            try:
                style = styles[style]
            except Exception:
                warning_message = "back to default style, wrong style entry, choose style from 0 to 23"
                print(warning_message)
                style = "default"
        pyplot.style.use(style)
        pyplot.figure(figsize=(20, 10))
        pyplot.title(f"{metric} ({len(training_history)} ephocs)")
        x = [i for i in range(0, len(training_history))]
        if metric == "all":
            for column in training_history.columns:
                if column != "Unnamed: 0":
                    pyplot.plot(x, training_history[column], label=str(column))
                    pyplot.legend()
                    pyplot.grid(True)
                    if save == True:
                        pyplot.savefig("plot.png")
        else:
            pyplot.plot(x, training_history[f"{metric}"], label="train")
            pyplot.plot(x, training_history[f"val_{metric}"], label="test")
            pyplot.legend()
            pyplot.grid(True)
            if save == True:
                pyplot.savefig("plot.png")


def main():
    from maaml.preprocessing import DataPreprocessor as dp

    processed = dp(dataset="UAHdataset", scaler=2)
    uahdataset = processed.preprocessed_dataset
    alias = processed.scaler_name
    model_build = DeepRCNModel()
    model_build.show()
    evaluation = Evaluator(
        model_build(),
        dataset=uahdataset,
        target_name="target",
        preprocessing_alias=alias,
        save_tag="test",
        input_shape=model_build.input_shape,
        cross_eval=True,
        epochs=2,
        verbose=1,
    )
    print("cross validation results: \n", evaluation.cross_evaluation)
    print("model name :", evaluation.model_name)


if __name__ == "__main__":
    main()



import pkg_resources
import pandas as pd
from sklearn.preprocessing import LabelEncoder, OneHotEncoder
from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
    MaxAbsScaler,
    RobustScaler,
    QuantileTransformer,
    PowerTransformer,
    Normalizer,
)
from maaml.utils import save_csv
import time


class DataPreprocessor:
    """A class for Data preprocessing specialized in time series data analysis from dataframes. includes these attributes: `raw_dataset`, `filtered_dataset` ,`numeric_dataset` ,`scaled_dataset` ,`scaler_name` ,`ml_dataset` ,`features` ,`target` ,`target_ohe` ,`preprocessed_dataset` ,
    `dl_dataset` and in the case of window stepping `windowed_dataset` ,`ml_dataset_w` ,`features_w` ,`target_w` ,`target_ohe_w` ,`preprocessed_dataset_w` ,`dl_dataset_w`.
    It also includes useful static methods a `uahdataset_loading` for loading the UAHdataset,`label_encoding` for encoding categorical data,`data_scaling` for scaling the data,`one_hot_encoding` for one hot encoding and `window_stepping` for window stepping.

    Args:
    * data_path (str, optional): The data file name in the working directory or the data file path with the file name used in case the dataset prarameter is not set. Defaults to `""`.
    * specific_data (str or int, optional): A parameter to define a specific grouping from the UAHdataset used in case the dataset prarameter is not set. Defaults to `None`.
    * target_name (str, optional): The name of the dataset target as a string. Defaults to `"target"`.
    * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features in columns and a target in one column with a name that match the target provided in the `target_name`. Defaults to `None`.
    * scaler (str, optional): selects the scaling technique used as integers from `"0"` to `"8"` passed as strings, or the name of the scaling technique such as `"minmax"` or `"normalizer"`. Defaults to no scaling with the value `"0"`.
    * droped_columns (list, optional): list of strings with the name of the columns to be removed or droped from the dataset after preprocessing. Defaults to `["Timestamp (seconds)"]`.
    * no_encoding_columns (list, optional): list of strings with the name of columns that will not be included in the label encoding process of cataegorical data. Defaults to `[]`.
    * no_scaling_columns (list, optional): list of strings with the name of columns not to be included in the data scaling. Defaults to `["target"]`.
    * window_size (int, optional): the size of the window in the case of window stepping the data, in case of `0` will not perform the window stepping. Defaults to `0`.
    * step (int, optional): The length of the step for window stepping, if smaller than `window_size` will result in overlapping windows, if equal to `window_size` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `0`.
    * window_transformation (bool, optional): in case of True applies the function in `window_transformation_function` parameter to the window. Defaults to `False`.
    * window_transformation_function (function, optional): A function to be applied to the window preferably a lambda function. Defaults to the mean value with: `lambda x:sum(x)/len(x)`.
    * from_csv (bool, optional): Specifies if the data loaded from data_path is a csv file or not. Defaults to `True`.
    * save_dataset (bool, optional): saves in a newly created directory under the working directory in the case of `True`, the preprocessed dataset with an ML specified and DL specified datasets, and windowed data for each case if window stepping is applied. Defaults to `False`.
    * save_tag (str, optional): add a custom tag to the name of the files to be saved in the case of save_dataset is `True`. Defaults to `"dataset"`.
    * verbose (int, optional): An integer of the verbosity of the process can be ``0`` or ``1``. Defaults to ``0``.
    """

    def __init__(
        self,
        data_path="",
        specific_data=None,
        target_name="target",
        dataset=None,
        scaler="0",
        droped_columns=["Timestamp (seconds)"],
        no_encoding_columns=[],
        no_scaling_columns=["target"],
        window_size=0,
        step=0,
        window_transformation=False,
        window_transformation_function=lambda x: sum(x) / len(x),
        from_csv=True,
        save_dataset=False,
        save_tag="dataset",
        verbose=0,
    ):
        """A constructor for the DataProcessor class

        Args:
            * data_path (str, optional): The data file name in the working directory or the data file path with the file name used in case the dataset prarameter is not set. Defaults to `""`.
            * specific_data (str or int, optional): A parameter to define a specific grouping from the UAHdataset used in case the dataset prarameter is not set. Defaults to `None`.
            * target_name (str, optional): The name of the dataset target as a string. Defaults to `"target"`.
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features in columns and a target in one column with a name that match the target provided in the `target_name`. Defaults to `None`.
            * scaler (str, optional): selects the scaling technique used as integers from `"0"` to `"8"` passed as strings, or the name of the scaling technique such as `"minmax"` or `"normalizer"`. Defaults to no scaling with the value `"0"`.
            * droped_columns (list, optional): list of strings with the name of the columns to be removed or droped from the dataset after preprocessing. Defaults to `["Timestamp (seconds)"]`.
            * no_encoding_columns (list, optional): list of strings with the name of columns that will not be included in the label encoding process of cataegorical data. Defaults to `[]`.
            * no_scaling_columns (list, optional): list of strings with the name of columns not to be included in the data scaling. Defaults to `["target"]`.
            * window_size (int, optional): the size of the window in the case of window stepping the data, in case of `0` will not perform the window stepping. Defaults to `0`.
            * step (int, optional): The length of the step for window stepping, if smaller than `window_size` will result in overlapping windows, if equal to `window_size` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `0`.
            * window_transformation (bool, optional): in case of True applies the function in `window_transformation_function` parameter to the window. Defaults to `False`.
            * window_transformation_function (function, optional): A function to be applied to the window preferably a lambda function. Defaults to the mean value with: `lambda x:sum(x)/len(x)`.
            * from_csv (bool, optional): Specifies if the data loaded from data_path is a csv file or not. Defaults to `True`.
            * save_dataset (bool, optional): saves in a newly created directory under the working directory in the case of `True`, the preprocessed dataset with an ML specified and DL specified datasets, and windowed data for each case if window stepping is applied. Defaults to `False`.
            * save_tag (str, optional): add a custom tag to the name of the files to be saved in the case of save_dataset is `True`. Defaults to `"dataset"`.
            * verbose (int, optional): An integer of the verbosity of the process can be ``0`` or ``1``. Defaults to ``0``.
        """
        start_time = time.perf_counter()
        if dataset is None or isinstance(dataset, str):
            if from_csv is True:
                if dataset in [
                    "UAHdataset",
                    "uahdataset",
                    "UAHDataset",
                    "UAHDATASET",
                    "uah",
                    "UAH",
                ]:
                    self.raw_dataset = self.uahdataset_loading(
                        data_path, specific=specific_data, verbose=verbose
                    )
                else:
                    try:
                        self.raw_dataset = pd.read_csv(data_path)
                        if verbose == 1:
                            print(
                                "Reading data from provided path to the data csv file"
                            )
                    except Exception:
                        print("\nError reading data, verify the provided path")
        elif dataset is not None:
            self.raw_dataset = dataset
            if verbose == 1:
                print("Reading from the dataset argement the provided dataframe")
        self.filtered_dataset = self.raw_dataset.drop(labels=droped_columns, axis=1)
        self.numeric_dataset = self.filtered_dataset.copy(deep=True)
        for column in self.numeric_dataset.columns:
            if (
                self.numeric_dataset.dtypes[column] != float
                and self.numeric_dataset.dtypes[column] != int
            ):
                if column in no_encoding_columns:
                    if verbose == 1:
                        print(
                            f"skipping \033[1m{column}\033[0m label encoding for being in the no_encoding_columns"
                        )
                else:
                    self.numeric_dataset = self.label_encoding(
                        self.numeric_dataset, target=column, verbose=verbose
                    )
        self.scaled_dataset, self.scaler_name = self.data_scaling(
            self.numeric_dataset,
            excluded_axis=no_scaling_columns,
            scaler=scaler,
            verbose=verbose,
        )
        self.ml_dataset = self.scaled_dataset
        self.features = self.ml_dataset.drop(target_name, axis=1)
        self.target = self.ml_dataset[target_name]
        self.target_ohe = self.one_hot_encoding(
            self.ml_dataset, target=target_name, verbose=verbose
        )
        self.preprocessed_dataset = self.ml_dataset.copy(deep=True)
        for i in self.target_ohe.columns:
            column_name = f"target {i}"
            self.preprocessed_dataset[column_name] = self.target_ohe[i]
        self.dl_dataset = self.preprocessed_dataset
        if window_size > 0:
            if verbose == 1:
                print(
                    "\n\033[1mThe window stepping can take some time depending on the dataset \033[0m"
                )
            self.windowed_dataset = self.ml_dataset.copy(deep=True)
            self.windowed_dataset = self.window_stepping(
                self.windowed_dataset,
                window_size=window_size,
                step=step,
                window_transformation=window_transformation,
                transformation_fn=window_transformation_function,
                verbose=verbose,
            )
            self.ml_dataset_w = self.windowed_dataset
            self.features_w = self.ml_dataset_w.drop(target_name, axis=1)
            if window_transformation == True:
                self.target_w = self.ml_dataset_w[target_name].round()
            else:
                self.target_w = self.ml_dataset_w[target_name]
            self.target_ohe_w = self.one_hot_encoding(
                self.target_w, target=target_name, verbose=verbose
            )
            self.preprocessed_dataset_w = self.ml_dataset_w.copy(deep=True)
            for i in self.target_ohe_w.columns:
                column_name = f"target {i}"
                self.preprocessed_dataset_w[column_name] = self.target_ohe_w[i]
            self.dl_dataset_w = self.preprocessed_dataset_w
        if save_dataset == True:
            PATH = "preprocessed_dataset"
            save_csv(self.ml_dataset, PATH, f"ml_{save_tag}", verbose=verbose)
            save_csv(self.dl_dataset, PATH, f"dl_{save_tag}", verbose=verbose)
            if window_size > 0:
                save_csv(
                    self.ml_dataset_w,
                    PATH,
                    f"ml_{save_tag}_w({window_size})_s({step})",
                    verbose=verbose,
                )
                save_csv(
                    self.dl_dataset_w,
                    PATH,
                    f"dl_{save_tag}_w({window_size})_s({step})",
                    verbose=verbose,
                )
        self.preprocessing_time = f"{(time.perf_counter() - start_time):.2f} (s)"
        exec_time = self.preprocessing_time.replace("(", "").replace(")", "")
        self.preprocessing_info = self.scaler_name + f"({exec_time})"

    @staticmethod
    def uahdataset_loading(path="", specific=None, verbose=1):
        """A static method for loading the uahdatset with various configirations.

        Args:
            * path (str, optional): The path to the dataset, if not provided, the internal dataset will be loaded. Defaults to "".
            * specific_data (str or int, optional): A parameter to define a specific grouping from the UAHdataset, if an integer used, the function will return the dataset of the driver matching that integer, if `""` or `"secondary road"` will return the data in the secondary road only, the same goes for `"0"` and `"motorway road"` returning the motorway road data only, if `None` the whole dataset is returned.the Defaults to `None`.
            * verbose (int, optional): An integer of the verbosity of the process can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: A UAHdataset grouping for time series data
        """
        if path == "":
            DATA_PATH = pkg_resources.resource_filename(
                __name__, "Datasets/UAH_dataset/dataset/UAHDataset.csv"
            )
            print(f"\nloading the internal \033[1mUAHDataset\033[0m from maaml\n")
            data = pd.read_csv(DATA_PATH)
        else:
            try:
                data = pd.read_csv(path)
                if verbose == 1:
                    print("\nUAHDataset read successfully\n")
            except Exception:
                print("\nERROR: bad path entry\nEmpty data variable returned")
                data = []
                return data
        if specific is None:
            data_info = "full data loaded successfully\n"
        elif str(specific) == "secondary road" or str(specific) == "":
            data = data.loc[data["road"] == "secondary"]
            data = data.drop("road", axis=1)
            data_info = "data of secondary road loaded successfully"
        elif str(specific) == "motorway road" or str(specific) == "0":
            data = data.loc[data["road"] == "motorway"]
            data = data.drop("road", axis=1)
            data_info = "data of motorway road loaded successfully"
        elif int(specific) < 7:
            data = data.loc[data["driver"] == int(specific)]
            data = data.drop("driver", axis=1)
            data_info = f"data of driver number {int(specific)} loaded successfully \n"
        else:
            print(
                "ERROR: wrong specific entry or specific entry does not exist\nEmpty data returned "
            )
            data = []
        if verbose == 1:
            print(data_info)
        return data

    @staticmethod
    def label_encoding(data, target, verbose=1):
        """A static method to to convert categorical data column to numeric data via label encoding.

        Args:
            * data (pandas.DataFrame or array or numpy.array): An array of data with a column of categorical data.
            * target ([str]): The name of the column to be converted.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: the data with the cateorical data column converted to numeric data.
        """
        encoder = LabelEncoder()
        df = pd.DataFrame(data)
        try:
            if verbose == 1:
                print(
                    f"encoding the \033[1m{target}\033[0m column. The target labels are: {data[target].unique()} "
                )
            df[target] = encoder.fit_transform(data[target])
            if verbose == 1:
                print(f"The target labels after encoding : {df[target].unique()}")
        except Exception:
            print(
                f"ERROR: the column name '{target}' is not available in data\nno label encoding realized for this target\n"
            )
        return data

    @staticmethod
    def data_scaling(data, excluded_axis=[], scaler="minmax", verbose=1):
        """A static method to scale the data using 8 diffrent scaling techniques or returning the raw data with all column values conveted to floats.

        Args:
            * data (pandas.DataFrame): A numeric dataset in a pandas.DataFrame format.
            * excluded_axis (list, optional): A list of column names for the columns to be excluded from the scaling process. Defaults to `[]`.
            * scaler (str, optional): selects the scaling technique used as integers from `"0"` to `"8"` passed as strings, or the name of the scaling technique such as `"RawData (no scaling)"` or `"normalizer"`. Defaults to `"minmax"`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * tuple: (pandas.DataFrame,str) A scaled pandas.DataFrame Data and the name of the scaling technique used as string.
        """
        scaled_df = data
        scaled_df = scaled_df.drop(excluded_axis, axis=1)
        columns_names_list = scaled_df.columns
        scaler = str(scaler)
        if scaler == "0" or scaler == "raw_data":
            scaler_name = "RawData (no scaling)"
            scaled_df = pd.DataFrame()
            for column in data.columns:
                scaled_df[column] = data[column].astype("float")
                scaled_df = scaled_df.reset_index(drop=True)
            scaled_df = scaled_df.fillna(0)
            if verbose == 1:
                print(f"data was not scaled, returned: {scaler_name}")
            return scaled_df, scaler_name
        elif scaler == "1" or scaler == "minmax":
            scalerfunction = MinMaxScaler()
            scaler_name = "MinMaxscaler"
        elif scaler == "2" or scaler == "standard":
            scalerfunction = StandardScaler()
            scaler_name = "Standardscaler"
        elif scaler == "3" or scaler == "maxabs":
            scalerfunction = MaxAbsScaler()
            scaler_name = "MaxAbsScaler"
        elif scaler == "4" or scaler == "robust":
            scalerfunction = RobustScaler()
            scaler_name = "RobustScaler"
        elif scaler == "5" or scaler == "quantile_normal":
            scalerfunction = QuantileTransformer(output_distribution="normal")
            scaler_name = "QuantileTransformer using normal distribution"
        elif scaler == "6" or scaler == "quantile_uniform":
            scalerfunction = QuantileTransformer(output_distribution="uniform")
            scaler_name = "QuantileTransformer using uniform distribution"
        elif scaler == "7" or scaler == "power_transform":
            scalerfunction = PowerTransformer(method="yeo-johnson")
            scaler_name = "PowerTransformer using the yeo-johnson method"
        elif scaler == "8" or scaler == "normalizer":
            scalerfunction = Normalizer()
            scaler_name = "Normalizer"
        else:
            print("\nERROR: wrong data entry or wrong scaler type\ninput data returned")
            scaler_name = "Worning : No scaling (something went wrong)"
            return data, scaler_name
        scaled_df = scalerfunction.fit_transform(scaled_df)
        scaled_df = pd.DataFrame(scaled_df, columns=columns_names_list)
        for i in excluded_axis:
            scaled_df[i] = data[i]
        scaled_df = scaled_df.fillna(0)
        if verbose == 1:
            print(f"data scaled with the {scaler_name}")
        return scaled_df, scaler_name

    @staticmethod
    def one_hot_encoding(data, target="target", verbose=1):
        """A static method to convert a single column to a number of columns corresponding to the number of unique values in that columns by One Hot encoding.
        Example:
            >>> df
            index              Timestamp              Speed              driver               road              target
            0                  7                  65.2                  1                  secondary                  normal
            1                  8                  68.5                  1                  secondary                  normal
            2                  9                  73.6                  1                  secondary                  normal
            3                 10                  80.2                  1                  secondary                  agressif
            4                 11                  90.9                  1                  secondary                  agressif
            >>> print(df[target].unique())
            ['normal' 'agressif']
            >>> df_ohe = one_hot_encoding(df,target="target",verbose=0)
            >>> df_ohe
                    0    1
            0      0.0  1.0
            1      0.0  1.0
            2      0.0  1.0
            3      1.0  0.0
            4      1.0  0.0

        Args:
            * data (pandas.DataFrame): A data array in pandas.DataFrame format.
            * target (str, optional): The name of the target column that is going to be one hot encoded. Defaults to `"target"`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: the target column converted to binary format in a pandas.DataFrame with the number of columns corresponds to the unique values of the target column.
        """
        encoder = OneHotEncoder()
        try:
            if verbose == 1:
                print(f"One Hot Encoder target: {data[target].unique()}")
            encoded = encoder.fit_transform(
                data[target].values.reshape(-1, 1)
            ).toarray()
        except Exception:
            try:
                if verbose == 1:
                    print(f"One Hot Encoder target: {data.unique()}")
                encoded = encoder.fit_transform(data.values.reshape(-1, 1)).toarray()
            except Exception:
                if verbose == 1:
                    print(
                        f"ERROR: target name '{target}' is not available in data\nNo One hot encoding realized"
                    )
                return data
        if verbose == 1:
            print(f"example of the target after One Hot encoding : {encoded[0]}")
        df = pd.DataFrame(encoded)
        return df

    @staticmethod
    def window_stepping(
        data=None,
        window_size=0,
        step=0,
        window_transformation=False,
        transformation_fn=lambda x: sum(x) / len(x),
        verbose=1,
    ):
        """A static method for window stepping a time series data.

        Args:
            * data (pandas.DataFrame, optional): A data array in pandas.DataFrame format. Defaults to `None`.
            * window_size (int, optional): the size of the window, in case of `0` will not perform the window stepping. Defaults to `0`.
            * step (int, optional): The length of the step, if smaller than `window_size` will result in overlapping windows, if equal to `window_size` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `0`.
            * window_transformation (bool, optional): in case of True applies the function in `window_transformation_function` parameter to the window. Defaults to `False`.
            * window_transformation_function (function, optional): A function to be applied to the window preferably a lambda function. Defaults to the mean value with: `lambda x:sum(x)/len(x)`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: A window stepped data in case the window was bigger than 0 or the entry dataframe in case window_size is equal to 0.
        """
        final_data = pd.DataFrame()
        if len(data) != 0:
            if window_size == 0:
                final_data = data
                if verbose == 1:
                    print("\nATTENTION: Entry data returned without window stepping")
                return final_data
            else:
                if verbose == 1:
                    if window_transformation is True:
                        print("\n\033[1mWindow transformation applied\033[0m")
                    else:
                        print(
                            f"\nwindow stepping applied with window size: {window_size} and step : {step}"
                        )
                for i in range(0, len(data) - 1, step):
                    window_segment = data[i : i + window_size]
                    if window_transformation is True:
                        window_segment = window_segment.apply(transformation_fn, axis=0)
                    final_data = final_data.append(window_segment, ignore_index=True)
        else:
            print("ERROR: Empty data entry")
        return final_data


if __name__ == "__main__":
    preprocessor = DataPreprocessor(
        dataset="UAHdataset",
        no_encoding_columns=[],
        scaler=2,
        window_size=60,
        step=10,
        window_transformation=True,
        window_transformation_function=lambda x: sum(x) / len(x),
        save_dataset=False,
        verbose=1,
    )
    print(f"\nthe raw dataset is: \n{preprocessor.raw_dataset}")
    print(f"\nthe dataset(after dropping columns) is\n{preprocessor.filtered_dataset}")
    print(f"the label encoded dataset: \n{preprocessor.numeric_dataset}")
    print(f"The used scaler is: {preprocessor.scaler_name}")
    print(f"\nthe scaled dataset is: \n{preprocessor.scaled_dataset}")
    print(f"\nthe dataset features are: \n{preprocessor.features}")
    print(f"\nthe dataset target column is: \n{preprocessor.target}")
    print(f"\nthe dataset one hot encoded target is: \n{preprocessor.target_ohe}")
    print(f"\nthe full preprocessed dataset is: \n{preprocessor.preprocessed_dataset}")
    print("\n ******* windowed data ******* \n")
    print(f"\nthe dataset windowed features are: \n{preprocessor.features_w}")
    print(f"\nthe dataset windowed target column is: \n{preprocessor.target_w}")
    print(
        f"\nthe dataset windowed one hot encoded target is: \n{preprocessor.target_ohe_w}"
    )
    print(
        f"\nthe full windowed preprocessed dataset is: \n{preprocessor.preprocessed_dataset_w}"
    )
    print(f"\nthe preprocessing time is : {preprocessor.preprocessing_time}")
    print(f"\npreprocessing info : {preprocessor.preprocessing_info}")



import pandas as pd
from maaml.utils import save_csv


class DataReader:
    """A class for reading data from a csv file. includes a `path` attribute and `data` attribute and a `__call__ ` method for calling an instance of the class to return the `data` attribute.

    Args:
        * path (str): The data file name in the working directory or the data file path with the file name.
        * header (int, optional):  The specification of the technique used to define the columns names: `None` in case of no columns names in the file, `0` in case that first row is the header. Defaults to `None`.
        * delimiter (str, optional): A string for the type of separation used in the csv file. Defaults to `" "`.
    """

    def __init__(self, path, header=None, delimiter=" "):
        """A constuctor for DataReader class

        Args:
        * path (str): The data file name in the working directory or the data file path with the file name.
        * header (int, optional):  The specification of the technique used to define the columns names: `None` in case of no columns names in the file, `0` in case that first row is the header. Defaults to `None`.
        * delimiter (str, optional): A string for the type of separation used in the csv file. Defaults to `" "`.
        """
        self.path = str(path)
        self.data = pd.read_table(path, header=header, delimiter=delimiter)

    def __call__(self):
        """A method for the class instance call

        Returns:
            * pandas.DataFrame: The read dataset from the file.
        """
        return self.data


class DataCleaner:
    """A class for data cleaning.

    Args:
    * data (pandas.DataFrame or array or numpy.array): A time series dataset.
    * merge_data (pandas.DataFrame or array or numpy.array, optional): A time series dataset to be merged with the data in the `data` parameter. Defaults to `None`.
    * window_size (int, optional): the size of the window in the case of window stepping the data, in case of `None` will not perform the window stepping. Defaults to `None`.
    * step (int, optional): The length of the step for window stepping,if `None` will not perform the window stepping, if smaller than `window_size` will result in overlapping windows, if equal to `window_size` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `None`.
    * drop_duplicates (bool, optional): if `True` removes the duplicate values using the Timestamp column as refrence. Defaults to `True`.
    * window_transformation (bool, optional): in case of True applies the function in `window_transformation_function` parameter to the window. Defaults to `False`.
    * window_transformation_function (function, optional): A function to be applied to the window preferably a lambda function. Defaults to the mean value with: `lambda x:sum(x)/len(x)`.
    * add_columns_dictionnary (dict, optional): A dictionnary of keys (column names) and values to be added to the array or the pandas.DataFrame, if set to `None` will skip adding data. Defaults to `None`.
    * save_dataset (bool, optional): in the case of `True` saves the dataset in a newly created directory under the working directory. Defaults to `False`.
    * save_tag (str, optional): add the name tag of the file to be saved in the case of save_dataset is `True`. Defaults to `"dataset"`.
    * timestamp_column (str, optional): the name of the column that has the timpestamps in seconds of the time series data. Defaults to `"Timestamp (seconds)"`.
    * verbose (int, optional): An integer of the verbosity of the process can be ``0`` or ``1``. Defaults to ``0``.
    """

    def __init__(
        self,
        data,
        merge_data=None,
        window_size: int = None,
        step: int = None,
        drop_duplicates=True,
        window_transformation=False,
        window_transformation_function=lambda x: sum(x) / len(x),
        add_columns_dictionnary: dict = None,
        save_dataset=False,
        save_tag="dataset",
        timestamp_column="Timestamp (seconds)",
        verbose=0,
    ):
        """A constructor for DataCleaner class.

        Args:
        * data (pandas.DataFrame or array or numpy.array): A time series dataset.
        * merge_data (pandas.DataFrame or array or numpy.array, optional): A time series dataset to be merged with the data in the `data` parameter. Defaults to `None`.
        * window_size (int, optional): the size of the window in the case of window stepping the data, in case of `None` will not perform the window stepping. Defaults to `None`.
        * step (int, optional): The length of the step for window stepping,if `None` will not perform the window stepping, if smaller than `window_size` will result in overlapping windows, if equal to `window_size` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `None`.
        * drop_duplicates (bool, optional): if `True` removes the duplicate values using the Timestamp column as reference. Defaults to `True`.
        * window_transformation (bool, optional): in case of True applies the function in `window_transformation_function` parameter to the window. Defaults to `False`.
        * window_transformation_function (function, optional): A function to be applied to the window preferably a lambda function. Defaults to the mean value with: `lambda x:sum(x)/len(x)`.
        * add_columns_dictionnary (dict, optional): A dictionnary of keys (column names) and values to be added to the array or the pandas.DataFrame, if set to `None` will skip adding data. Defaults to `None`.
        * save_dataset (bool, optional): in the case of `True` saves the dataset in a newly created directory under the working directory. Defaults to `False`.
        * save_tag (str, optional): add the name tag of the file to be saved in the case of save_dataset is `True`. Defaults to `"dataset"`.
        * timestamp_column (str, optional): the name of the column that has the timpestamps in seconds of the time series data. Defaults to `"Timestamp (seconds)"`.
        * verbose (int, optional): An integer of the verbosity of the process can be ``0`` or ``1``. Defaults to ``0``.
        """
        self.raw_data = data
        if drop_duplicates is True:
            self.filtered_data = data.drop_duplicates(subset=[timestamp_column])
        else:
            self.filtered_data = self.raw_data
            if verbose == 1:
                print(f"No duplicates droped, filtered_data is raw_data")
        if (window_size is None) and (step is None):
            self.windowed_data = self.filtered_data
            if verbose == 1:
                print(f"No window stepping,windowed_data is filtered_data")
        else:
            self.windowed_data = self.window_stepping(
                self.filtered_data,
                window_size=window_size,
                step=step,
                window_transformation=window_transformation,
                transformation_fn=window_transformation_function,
                verbose=verbose,
            )
        if merge_data is not None:
            self.merged_data = self.dataframes_merging(
                self.windowed_data,
                merge_data,
                timestamp_column=timestamp_column,
                drop_duplicates=drop_duplicates,
                verbose=verbose,
            )
        else:
            self.merged_data = self.windowed_data
            if verbose == 1:
                print(f"No data merging,merged_data is windowed_data")
        self.interpolated_data = self.data_interpolating(
            self.merged_data, timestamp_column=timestamp_column, verbose=verbose
        )
        self.dataset = self.removing_incomplete_raws(
            self.interpolated_data, verbose=verbose
        )
        if add_columns_dictionnary is not None:
            self.dataset = self.column_adding(
                self.dataset,
                add_columns_dictionnary=add_columns_dictionnary,
                verbose=verbose,
            )
        if save_dataset == True:
            PATH = "cleaned_dataset"
            save_csv(self.dataset, PATH, save_tag, verbose=verbose)

    def __call__(self):
        """A method for the class instance call

        Returns:
            * pandas.DataFrame: The cleaned dataset.
        """
        return self.dataset

    @staticmethod
    def window_stepping(
        data=None,
        window_size: int = None,
        step: int = None,
        window_transformation=False,
        transformation_fn=lambda x: sum(x) / len(x),
        verbose=1,
    ):
        """A static method for window stepping a time series data.

        Args:
            * data (pandas.DataFrame, optional): A data array in pandas.DataFrame format. Defaults to `None`.
            * window_size (int, optional): the size of the window, in case of `None` will not perform the window stepping. Defaults to `None`.
            * step (int, optional): The length of the step,if `None` will not perform the window stepping, if smaller than `window_size` will result in overlapping windows, if equal to `window_size` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `None`.
            * window_transformation (bool, optional): in case of True applies the function in `window_transformation_function` parameter to the window. Defaults to `False`.
            * window_transformation_function (function, optional): A function to be applied to the window preferably a lambda function. Defaults to the mean value with: `lambda x:sum(x)/len(x)`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: A window stepped data in case the window was bigger than 0 or the entry dataframe in case window_size is equal to 0.
        """

        if len(data) != 0:
            if (window_size is not None) and (step is not None):
                if window_size == 0 or step == 0:
                    if verbose == 1:
                        print(
                            "\nATTENTION: No window stepping,one of window_size or step are set to 0"
                        )
                    return data
                else:
                    final_data = pd.DataFrame()
                    if verbose == 1:
                        if window_transformation is True:
                            print("\n\033[1mWindow transformation applied\033[0m")
                        else:
                            print(
                                f"\nwindow stepping applied with window size: {window_size} and step : {step}"
                            )
                    for i in range(0, len(data) - 1, step):
                        window_segment = data[i : i + window_size]
                        if window_transformation is True:
                            window_segment = window_segment.apply(
                                transformation_fn, axis=0
                            )
                        final_data = final_data.append(
                            window_segment, ignore_index=True
                        )
            else:
                return data
        else:
            print("ERROR: Empty data entry")
        return final_data

    @staticmethod
    def dataframes_merging(
        data=None,
        new_data=None,
        timestamp_column="Timestamp (seconds)",
        drop_duplicates=True,
        verbose=1,
    ):
        """A static method for merging two dataframes.

        Args:
            * data (pandas.DataFrame, optional): A time series dataframe. Defaults to `None`.
            * new_data (pandas.DataFrame or array or numpy.array, optional): A time series dataset to be merged with the data in the `data` parameter. Defaults to `None`.
            * timestamp_column (str, optional): The name of the timestamp column to be used as reference for the merge operation. Defaults to `"Timestamp (seconds)"`.
            * drop_duplicates (bool, optional): if `True` removes the duplicate values  in  both dataframes using the Timestamp column as reference before the merge operation. Defaults to `True`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: a pandas.Dataframe with the two orginal dataframes merged using `timestamp_column` as reference.
        """
        try:
            while data.dtypes[timestamp_column] != "int64":
                if verbose == 1:
                    print(
                        "\nWarning: data Timestamp type is: ",
                        data.dtypes[timestamp_column],
                        "\n",
                    )
                data = data.astype({timestamp_column: "int"})
                if verbose == 1:
                    print(
                        "data timestamp type changed to : ",
                        data.dtypes[timestamp_column],
                        "\n",
                    )
            while new_data.dtypes[timestamp_column] != "int64":
                if verbose == 1:
                    print(
                        "Warning: new_data Timestamp type is: ",
                        data.dtypes[timestamp_column],
                        "\n",
                    )
                new_data = new_data.astype({timestamp_column: "int"})
                if verbose == 1:
                    print(
                        "new_data timestamp type changed to : ",
                        new_data.dtypes[timestamp_column],
                        "\n",
                    )
            if drop_duplicates is True:
                data = data.drop_duplicates([timestamp_column])
                new_data = new_data.drop_duplicates([timestamp_column])
            data_merged = data.set_index(timestamp_column).join(
                new_data.set_index(timestamp_column)
            )
            data_merged = data_merged.reset_index()
            if verbose == 1:
                print(f"Shape of the megred data: {data_merged.shape}\n")
                print("\033[1m", "******* DATA SUCCESSFULLY MERGED *******", "\033[0m")
        except Exception:
            print(
                "ERROR: empty data entries or one data entry or both do not have Timestamp column, \nplease renter your two dataframes and check their columns before entry "
            )
            print("\nEmpty data returned")
            data_merged = None
        return data_merged

    @staticmethod
    def data_interpolating(
        data=None, timestamp_column="Timestamp (seconds)", verbose=1
    ):
        """A static method for data interpolation.

        Args:
            * data (pandas.DataFrame, optional): A time series dataframe with missing values. Defaults to `None`.
            * timestamp_columns (str, optional): The name of the timestamp column for the time series data (to ignore this column in the interpolation). Defaults to `"Timestamp (seconds)"`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: A pandas.DataFrame with filled missing values.
        """
        try:
            if verbose == 1:
                print(
                    f"\n    State before interpolation    \nCOLUMNS                   NUMBER OF RAWS WITH MISSING DATA\n{data.isnull().sum()}\n"
                )
            if data.isnull().values.any() == True:
                if verbose == 1:
                    print("\n       Executing interpolation     \n")
                missing_values = data.drop(timestamp_column, axis=1)
                missing_values = missing_values.interpolate(method="cubic", limit=3)
                data[missing_values.columns] = missing_values
                data_interpolated = data
                if verbose == 1:
                    print(
                        f"\n    State after interpolation    \nCOLUMNS                   NUMBER OF RAWS WITH MISSING DATA\n{data_interpolated.isnull().sum()}\n"
                    )
            else:
                data_interpolated = data
                if verbose == 1:
                    print("\n   Interpolation not needed    \n")
        except Exception:
            data_interpolated = None
            print(
                f"\nERROR: empty data entry or non dataframe type\nEmpty data returned"
            )
        return data_interpolated

    @staticmethod
    def removing_incomplete_raws(data=None, verbose=1):
        """A static method for removing all Nan or Na values.

        Args:
            * data (pandas.DataFrame, optional): A dataframe. Defaults to `None`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``1``.

        Returns:
            * pandas.DataFrame: a dataframe with no missing or Nan/ Na/ None values.
        """
        try:
            if verbose == 1:
                print(
                    f"\n    Data count before removing any rows :     \n{data.count()}"
                )
                print(
                    "\nis there any missing data values? :",
                    "\033[1m",
                    data.isnull().values.any(),
                    "\033[0m",
                )
            data = data.dropna()
            data = data.reset_index(drop=True)
            if verbose == 1:
                print(f"\n  Final Data count :     \n{data.count()}")
                print(
                    "\nis there any missing data values? :",
                    "\033[1m",
                    data.isnull().values.any(),
                    "\033[0m",
                )
        except Exception:
            print(
                "ERROR: empty data entry or non dataframe type, please enter your data dataframe\nEmpty data returned"
            )
            data = []
        return data

    @staticmethod
    def column_adding(
        data,
        add_columns_dictionnary: dict = None,
        column_name: str = None,
        value: str = None,
        verbose=0,
    ):
        """A static method for adding columns into a pandas.DataFrame.

        Args:
            * data (pandas.DataFrame): A dataframe. Defaults to `None`.
            * add_columns_dictionnary (dict, optional): A dictionnary of keys (column names) and corresponding values to be added, if set to `None` will check `column_name` and `value` if one of them is `None` will skip adding data. Defaults to `None`.
            * column_name (str, optional): the name of the new column. Defaults to `None`.
            * value (Any, optional): The value of the new column. Defaults to `None`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``0``.

        Returns:
            [type]: [description]
        """
        if add_columns_dictionnary is None:
            if column_name is not None and value is not None:
                data[column_name] = value
            else:
                data = data
                if verbose == 1:
                    print("\n    No columns added    \n")
        else:
            for key in add_columns_dictionnary:
                try:
                    data[key] = add_columns_dictionnary[key]
                    if verbose == 1:
                        print(f"\nThe '{key}' column was added successfully   \n")
                except ValueError:
                    print(
                        f"\nAbort column adding operation : The '{key}' column length is {len(add_columns_dictionnary[key])}, while the data length {len(data)}\n"
                    )
                    break
        return data


if __name__ == "__main__":
    DATA_DIR_PATH = "/run/media/najem/34b207a8-0f0c-4398-bba2-f31339727706/home/stock/The_stock/dev & datasets/PhD/datasets/UAH-DRIVESET-v1/"
    my_dict = {
        "Timestamp (seconds)": [4, 5, 6, 7, 8, 9, 10, 11, 12, 13],
        "speed": [40, 45, 60, 62, 70, 75, 80, None, 72, 70],
        "loc": [3, 4, 7, 10, None, 15, 17, 20, 24, 27],
        "driver": ["D1", "D1", "D1", "D1", "D2", "D4", None, "D2", "D1", "D5"],
        "target": [
            "normal",
            "normal",
            "normal",
            "agressif",
            "agressif",
            "drowsy",
            "normal",
            "normal",
            "normal",
            "drowsy",
        ],
    }
    data = pd.DataFrame(my_dict)
    cleaning = DataCleaner(
        data,
        drop_duplicates=False,
        add_columns_dictionnary={"axis": [12, 4, 5, 7, 5, 8, 2, 5, 4]},
        save_dataset=True,
        verbose=1,
    )
    print(cleaning.raw_data)
    print(cleaning.dataset)



from sklearn.tree import DecisionTreeClassifier, ExtraTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn import svm
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import ShuffleSplit, train_test_split
from sklearn.preprocessing import label_binarize
from sklearn.metrics import (
    accuracy_score,
    precision_score,
    recall_score,
    f1_score,
    cohen_kappa_score,
    roc_auc_score,
    classification_report,
)
import numpy as np
import time
from maaml.utils import save_csv, dict_transpose


class Evaluator:
    """A class to evaluate a dataset with 9 different machine learning models and saves the results in the working directory. includes an `evaluation` attribute, a `feature_importance_ranks` attribute in case the model allows for feature importance extraction, a `model_name` attribute, a `model` attribute and the `save_tag` attribute.
    It also includes useful static methods a `model_building` that produces a machine learning model class, a `model_evaluating` that does not include cross validation in the evaluation and a `feature_importance_ranking` that produce the `feature_importance_ranks`

    Args:
        * model_name (str or, optional): The choice of machine learning model can be an integer between `"0"` and `"9"` passed as a string or a name from the 9 available models such as `"LogisticRegression"`. Defaults to `"4"` which is `"ExtraTrees"` the best performing model in the most case.
        * parameter (str, optional): The standard parameter passed to the model as a string can be for exemple `"1000"` for the number of estimators to `"RandomForest"` or `"ExtraTrees"` or `"lbfgs"` the solver for `"LogisticRegression"`. Defaults to `None`.
        * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
        * target (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
        * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features in columns and a classfication target in one column. Defaults to `None`.
        * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
        * nb_splits (int, optional): The number of splits in case of cross validation. Defaults to `5`.
        * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
        * full_eval (bool, optional): A full evaluation with all the available models in the case of `True` or the model chosen in the `model_name` parameter in the case of `False`. Defaults to `False`.
        * save_eval (bool, optional): saves the evaluation results in a new directory under the working directory in case of `True` and does not save the evaluation results in the case of `False`. Defaults to `False`.
        * save_tag (str, optional): The tag given to the created directory and the saved evaluation result, important in case of mutiple evaluation in the same directory to not overwite exiting results. Defaults to `None`.
        * preprocessing_alias (str, optional): The name for the applied dataset preprocessing as a string that is going to be displayed in the evaluation results. Defaults to `None`.
        * verbose (int, optional): An integer of the verbosity of the evaluation can be ``0`` or ``1``. Defaults to ``0``.
    """

    def __init__(
        self,
        model_name="4",
        parameter=None,
        features=None,
        target=None,
        dataset=None,
        target_name="target",
        nb_splits=5,
        test_size=0.3,
        full_eval=False,
        save_eval=False,
        save_tag=None,
        preprocessing_alias=None,
        verbose=0,
    ):
        """The constructor of the Evaluator class.

        Args:
            * model_name (str or, optional): The choice of machine learning model can be an integer between `"0"` and `"9"` passed as a string or a name from the 9 available models such as `"LogisticRegression"`. Defaults to `"4"` which is `"ExtraTrees"` the best performing model in the most case.
            * parameter (str, optional): The standard parameter passed to the model as a string can be for exemple `"1000"` for the number of estimators to `"RandomForest"` or `"ExtraTrees"` or `"lbfgs"` the solver for `"LogisticRegression"`. Defaults to `None`.
            * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
            * target (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features in columns and a classfication target in one column. Defaults to `None`.
            * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
            * nb_splits (int, optional): The number of splits in case of cross validation. Defaults to `5`.
            * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
            * full_eval (bool, optional): A full evaluation with all the available models in the case of `True` or the model chosen in the `model_name` parameter in the case of `False`. Defaults to `False`.
            * save_eval (bool, optional): saves the evaluation results in a newly created directory under the working directory in case of `True`. Defaults to `False`.
            * save_tag (str, optional): The tag given to the created directory and the saved evaluation result, important in case of mutiple evaluation in the same directory to not overwite exiting results. Defaults to `None`.
            * preprocessing_alias (str, optional): The name for the applied dataset preprocessing as a string that is going to be displayed in the evaluation results. Defaults to `None`.
            * verbose (int, optional): An integer of the verbosity of the evaluation can be ``0`` or ``1``. Defaults to ``0``.
        """
        self.save_tag = "" if save_tag is None or save_tag == "" else f"_{save_tag}"
        PATH = f"ML_EVALUATION{self.save_tag}"
        if preprocessing_alias is not None:
            self.save_tag = f"_{preprocessing_alias}" + self.save_tag
        else:
            preprocessing_alias = ""
        self.model = i = 1
        self.evaluation = {}
        while True:
            if full_eval is False:
                self.model = self.model_building(model_name, parameter, verbose)
            elif full_eval is True:
                try:
                    self.model = self.model_building(i, parameter, verbose)
                    i += 1
                except ValueError:
                    print("full evaluation complete")
                    self.evaluation = dict_transpose(self.evaluation)
                    if save_eval is True:
                        self.tag = f"full_evaluation{self.save_tag}"
                        save_csv(self.evaluation, PATH, self.tag, verbose)
                    break

            if "SVC" in str(self.model):
                self.model_name = str(self.model).replace("SVC", "SVMClassifier")
            self.model_name = str(self.model).replace("()", "")

            cross_evaluation = self.model_cross_validating(
                features,
                target,
                dataset,
                target_name,
                nb_splits,
                test_size,
                preprocessing_alias,
                verbose,
            )

            if not self.evaluation:
                self.evaluation = cross_evaluation
            else:
                for key in cross_evaluation:
                    self.evaluation[key].append(*cross_evaluation[key])
            if full_eval is False:
                self.evaluation = dict_transpose(self.evaluation)
                if save_eval is True:
                    self.tag = f"{self.model_name}{self.save_tag}_evaluation"
                    save_csv(self.evaluation, PATH, self.tag, verbose)

            try:
                self.feature_importance_ranks = self.features_importance_ranking(
                    self.model,
                    dataset,
                    target_name,
                    features,
                    target,
                    test_size,
                    verbose,
                )
            except AttributeError:
                self.feature_importance_ranks = None
                if verbose == 1:
                    print(
                        f"The {str(self.model)} does not allow the extraction of feature importance ranks\nSkipping action"
                    )
            if full_eval is False:
                break

    @staticmethod
    def model_building(model_name="4", parameter=None, verbose=0):
        """A static method to create a non trained machine learning model class.

        Args:
            * model_name (str or, optional): The choice of machine learning model can be an integer between `"0"` and `"9"` passed as a string or a name from the 9 available models such as `"LogisticRegression"`. Defaults to `"4"` which is `"ExtraTrees"` the best performing model in the most case.
            * parameter (str, optional): The standard parameter passed to the model as a string can be for exemple `"1000"` for the number of estimators to `"RandomForest"` or `"ExtraTrees"` or `"lbfgs"` the solver for `"LogisticRegression"`. Defaults to `None`.
            * verbose (int, optional): An integer of the verbosity of the model selection process can be ``0`` or ``1``. Defaults to ``0``.

        Returns:
            * class: a machine learning model class or a `"No model"` string in case of wrong selection of the `model_name` parameter.
        """
        model_name = str(model_name)
        if model_name == "1" or model_name == "DecisionTree":
            model = DecisionTreeClassifier()
        elif model_name == "2" or model_name == "RandomForest":
            if parameter is not None:
                parameter = int(parameter)
                model = RandomForestClassifier(n_estimators=parameter)
            else:
                model = RandomForestClassifier()
        elif model_name == "3" or model_name == "ExtraTree":
            model = ExtraTreeClassifier()
        elif model_name == "4" or model_name == "ExtraTrees":
            if parameter is not None:
                parameter = int(parameter)
                model = ExtraTreesClassifier(n_estimators=parameter)
            else:
                model = ExtraTreesClassifier()
        elif model_name == "5" or model_name == "KNeighbors":
            if parameter is not None:
                parameter = int(parameter)
                model = KNeighborsClassifier(n_neighbors=parameter)
            else:
                model = KNeighborsClassifier()
        elif model_name == "6" or model_name == "GaussianNB":
            model = GaussianNB()
        elif model_name == "7" or model_name == "SVM":
            if parameter is not None:
                parameter = str(parameter)
                model = svm.SVC(gamma=parameter)
            else:
                model = svm.SVC()
        elif model_name == "8" or model_name == "LogisticRegression":
            if parameter is not None:
                parameter = str(parameter)
                model = LogisticRegression(
                    solver=parameter, multi_class="auto", max_iter=1000
                )
            else:
                model = LogisticRegression(multi_class="auto", max_iter=1000)
        elif model_name == "9" or model_name == "MLPClassifier":
            if parameter is not None:
                parameter = int(parameter)
                model = MLPClassifier(max_iter=parameter)
            else:
                model = MLPClassifier()
        else:
            error_message = "ERROR:wrong entry,you have 9 different classifiers, you could choose by number or by name"
            print(error_message)
            model = "No model"
        if verbose == 1:
            print(f"\nThe {str(model)} is selected")
        return model

    def model_cross_validating(
        self,
        features=None,
        target=None,
        dataset=None,
        target_name="target",
        nb_splits=5,
        test_size=0.3,
        preprocessing_alias=None,
        verbose=0,
    ):
        """A method for cross evaluating and training of a machine learning model class.

        Args:
            * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
            * target (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features in columns and a classfication target in one column. Defaults to `None`.
            * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
            * nb_splits (int, optional): The number of splits in case of cross validation. Defaults to `5`.
            * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
            * preprocessing_alias (str, optional): The name for the applied dataset preprocessing as a string that is going to be displayed in the evaluation results. Defaults to `None`.
            * verbose (int, optional): An integer of the verbosity of the evaluation can be ``0`` or ``1``. Defaults to ``0``.

        Returns:
            dict: `cv_scores` the result cross validation evaluation scores.
        """
        start_time = time.perf_counter()
        if dataset is not None:
            X, Y = dataset.drop(target_name, axis=1), dataset[target_name]
        elif features is not None and target is not None:
            X, Y = features, target
        else:
            error_message = "ERROR: please enter a dataset with a target_name or you can enter features and target"
            print(error_message)
            return
        cv = ShuffleSplit(n_splits=nb_splits, test_size=test_size, random_state=10)
        acc_scores, pres_scores, rec_scores, f1, cokap_scores, roc_auc_scores = (
            [],
            [],
            [],
            [],
            [],
            [],
        )
        cv_scores = {
            "MLclassifier": [],
            "preprocessing": [],
            "execution time": [],
            "accuracy": [],
            "precision": [],
            "recall": [],
            "F1": [],
            "cohen_kappa": [],
            "roc_auc": [],
        }
        for train, test in cv.split(X, Y):
            classes = Y.unique()
            y_testb = label_binarize(Y[test], classes=classes)
            Y_values = Y.values
            Y_reshaped = Y_values.reshape(-1, 1)
            model = self.model
            pred = model.fit(X.loc[train], Y_values[train]).predict(X.loc[test])
            pred_reshaped = pred.reshape(-1, 1)
            acc_scores.append(accuracy_score(Y_values[test], pred, normalize=True))
            pres_scores.append(precision_score(Y_values[test], pred, average="macro"))
            rec_scores.append(recall_score(Y_values[test], pred, average="macro"))
            f1.append(f1_score(Y_values[test], pred, average="macro"))
            cokap_scores.append(cohen_kappa_score(Y_reshaped[test], pred_reshaped))
            roc_auc_scores.append(roc_auc_score(y_testb, pred_reshaped))
        end_time = time.perf_counter()
        cv_scores["MLclassifier"].append(self.model_name)
        cv_scores["preprocessing"].append(preprocessing_alias)
        cv_scores["execution time"].append(
            f"{((end_time-start_time) / nb_splits): .2f} (s)"
        )
        cv_scores["accuracy"].append(
            f"{np.mean(acc_scores):.4%} (+/- {np.std(acc_scores):.4%})"
        )
        cv_scores["precision"].append(
            f"{np.mean(pres_scores):.4%} (+/- {np.std(pres_scores):.4%})"
        )
        cv_scores["recall"].append(
            f"{np.mean(rec_scores):.4%} (+/- {np.std(rec_scores):.4%})"
        )
        cv_scores["F1"].append(f"{np.mean(f1):.4%} (+/- {np.std(f1):.4%})")
        cv_scores["cohen_kappa"].append(
            f"{np.mean(cokap_scores):.4%} (+/- {np.std(cokap_scores):.4%})"
        )
        cv_scores["roc_auc"].append(
            f"{np.mean(roc_auc_scores):.4%} (+/- {np.std(roc_auc_scores):.4%})"
        )
        if verbose == 1:
            print("\n\033[1mCross validation results: \033[0m")
            for i, v in cv_scores.items():
                print(f"{i}: {v[0]}")
        if verbose == 2:
            print(f"\nAccuracy evaluation for the separate splits:\n{acc_scores}")
            print(f"\nPrecision evaluation for the separate splits:\n{pres_scores}")
            print(f"\nRecall evaluation for the separate splits:\n{rec_scores}")
            print(f"\nF1 evaluation for the separate splits:\n{f1}")
            print(f"\nCohen_kappa evaluation for the separate splits:\n{cokap_scores}")
            print(f"\nRoc_Auc evaluation for the separate splits:\n{roc_auc_scores}")
        return cv_scores

    @staticmethod
    def features_importance_ranking(
        model=None,
        dataset=None,
        target_name="target",
        features=None,
        target=None,
        test_size=0.3,
        verbose=0,
    ):
        """A static method to evaluate the features importance and ranks them.

        Args:
            * model (class, optional): A machine learning model class that allowes for feature importance extraction, if left to `None` uses ExtraTreesClassifier as the default for the feature_importance_ranking. Defaults to `None`.
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features in columns and a classfication target in one column. Defaults to `None`.
            * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
            * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
            * target (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
            * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
            * verbose (int, optional): An integer of the verbosity of the process, can be ``0`` or ``1``. Defaults to ``0``.

        Returns:
            * [pandas.DataFrame class]: The features names in one column and the importance percentage of every feature in another column.
        """
        if dataset is not None and target_name is not None:
            x = dataset.drop(target_name, axis=1)
            y = dataset[target_name].values
        elif dataset is None:
            x = features
            y = target.values
        X_train, X_test, y_train, y_test = train_test_split(
            x, y, test_size=test_size, random_state=10
        )
        if model is None:
            learner = ExtraTreesClassifier()
            model = str(learner)
            print(f"The default Classifier for feature importance ranking is {model}")
        else:
            learner = model
        learner = learner.fit(X_train, y_train)
        pred = learner.predict(X_test)

        if verbose == 1:
            f1score = f1_score(y_test, pred, average="macro")
            print(
                f"\nTrying to use {str(model)} for the feature ranking with an F1 score of : {f1score*100: .2f}%\n"
            )
        importances = learner.feature_importances_
        ranks = x.T.drop(x.index, axis=1)
        ranks["importance %"] = importances * 100
        ranks = ranks.sort_values("importance %")[::-1]
        if verbose == 1:
            print(
                f"The {len(ranks)} features importance is ranked successfully using the {str(learner)}"
            )
        return ranks

    @staticmethod
    def model_evaluating(
        model=None,
        dataset=None,
        target_name="target",
        features=None,
        targets=None,
        test_size=0.3,
        verbose=0,
    ):
        """A static method for evaluating a model and producing a classification report.

        Args:
            * model (class, optional): A machine learning model class, if left to `None` uses ExtraTreesClassifier as the default evaluator. Defaults to `None`.
            * dataset (pandas.DataFrame or array or numpy.array, optional): A dataset that includes features in columns and a classfication target in one column. Defaults to `None`.
            * target_name (str, optional): The name of the classification target as a string. Defaults to `"target"`.
            * features (pandas.DataFrame or array or numpy.array, optional): The features of the dataset in columns for the case that the dataset parameter is not provided. Defaults to `None`.
            * target (pandas.Series or array or numpy.array, optional): The classfication target in one column in the case that the dataset parameter is not provided . Defaults to `None`.
            * test_size (float, optional): The percentage of the test sample size as a float from the full dataset represented as `1` . Defaults to `0.3`.
            * verbose (int, optional): An integer of the verbosity of the evaluation, can be ``0`` or ``1``. Defaults to ``0``.

        Returns:
            * str: A string in a table structure for the classfication report.
        """
        if dataset is not None and target_name is not None:
            x = dataset.drop(target_name, axis=1)
            y = dataset[target_name].values
        elif dataset is None:
            x = features
            y = targets.values
        else:
            print("No data is provided for the evaluation")
        X_train, X_test, y_train, y_test = train_test_split(
            x, y, test_size=test_size, random_state=10
        )
        if model is None:
            learner = ExtraTreesClassifier()
            model = str(learner)
            print(
                f"You did not provide a classifier, the default Classifier is {model}"
            )
        else:
            learner = model
        learner = learner.fit(X_train, y_train)
        pred = learner.predict(X_test)
        results = classification_report(y_test, pred)
        if verbose == 1:
            print(results)
        return results


def main():
    from maaml.preprocessing import DataPreprocessor as dp

    processed = dp(dataset="UAHdataset", scaler=2)
    uahdataset = processed.ml_dataset
    alias = processed.scaler_name
    ml_evaluation = Evaluator(
        3,
        dataset=uahdataset,
        verbose=1,
        preprocessing_alias=alias,
        full_eval=False,
        save_eval=True,
    )
    print(f"feature importance :\n{ml_evaluation.feature_importance_ranks}")


if __name__ == "__main__":
    main()



import pandas as pd
from maaml.utils import save_csv, FileScraper, pattern_search, read_csv, save_parquet
from maaml.cleaning import DataCleaner


class UahDatasetPathFinder:
    """A class for generating a path for a specific file in the UahDataset. includes an attribute `file_path` and a method `numeric_to_string_from_list` for converting a nemeric value to the a corresponding indexed string from a list and a `__call__ ` method for calling an instance of the class to return the `file_path` attribute.

        Args:
        * dataset_dir (str): The path to the parent directory of the UahDataset.
        * driver (str or int): The number of the driver, can be an integer from `1` to `6` or a string that ends with integer in the same range such as `"driver1"` or `"D6"`.
        * state (str or int): The state of the driving session, can be an integer from `1` to `3` or a string such as `"normal"`,`"aggressive"` or `"drowsy"` not case sensitive.
        * road (str or int): The type of road in the driving session, can be an integer from `1` to `2` or a string such as `"motorway"` or `"secondary"` not case sensitive.
        * data_type (str or int): The type of data or the file of data chosen, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
        * standard (bool, optional): if `True` gives you the first session and if `False` gives the second session, of course for both cases it depends on the availability of the data. Defaults to `True`.
        * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``0``.

    Raises:
        * ValueError: In the case of wrong argement type or entry, with a description for the reason.
    """

    SENSOR_FILES = [
        "RAW_GPS.txt",
        "RAW_ACCELEROMETERS.txt",
        "PROC_LANE_DETECTION.txt",
        "PROC_VEHICLE_DETECTION.txt",
        "PROC_OPENSTREETMAP_DATA.txt",
        "EVENTS_LIST_LANE_CHANGES.txt",
        "EVENTS_INERTIAL.txt",
        "SEMANTIC_FINAL.txt",
        "SEMANTIC_ONLINE.txt",
    ]
    STATES = ["NORMAL", "AGGRESSIVE", "DROWSY"]
    ROADS = ["MOTORWAY", "SECONDARY"]

    def __init__(
        self, dataset_dir, driver, state, road, data_type, standard=True, verbose=0
    ):
        """A constructor for the UahDatasetPathFinder class.

            Args:
            * dataset_dir (str): The path to the parent directory of the UahDataset.
            * driver (str or int): The number of the driver, can be an integer from `1` to `6` or a string that ends with integer in the same range such as `"driver1"` or `"D6"`.
            * state (str or int): The state of the driving session, can be an integer from `1` to `3` or a string such as `"normal"`,`"aggressive"` or `"drowsy"` not case sensitive.
            * road (str or int): The type of road in the driving session, can be an integer from `1` to `2` or a string such as `"motorway"` or `"secondary"` not case sensitive.
            * data_type (str or int): The type of data or the file of data chosen, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
            * standard (bool, optional): if `True` gives you the first session and if `False` gives the second session, of course for both cases it depends on the availability of the data. Defaults to `True`.
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1`` or ``2``. Defaults to ``0``.

        Raises:
            * ValueError: In the case of wrong argement type or entry, with a description for the reason.
        """
        file_error_msg = (
            "please give a correct data_type from the files available for the dataset"
        )
        numeric_file_error_msg = (
            "please give the correct number of the file, only 9 file available"
        )
        data_type = self.numeric_to_string_from_list(
            data_type, self.SENSOR_FILES, numeric_file_error_msg
        ).replace(".txt", "")
        file_name = pattern_search(data_type.upper(), self.SENSOR_FILES, file_error_msg)
        if len(file_name) > 1:
            raise ValueError(
                f"There is more than one data_type with that entry:\n{file_name}\nplease be more specific."
            )
        else:
            for file in file_name:
                self.file = file
                if verbose == 1:
                    print(
                        f"Searching '{file}' file in all directories and sub_directories.. "
                    )

            scraper_verbose = 1 if verbose > 1 else verbose
            self.path_list = set(
                FileScraper(dataset_dir, file_name, verbose=scraper_verbose).path_list,
            )
            driver = (
                driver
                if isinstance(driver, int) or driver.isnumeric()
                else int(driver[-1])
            )
            driver_error_msg = (
                "please give a correct driver number, only 6 drivers available."
            )
            driver_name = f"D{driver}"
            self.driver = driver_name
            driver_paths = pattern_search(driver_name, self.path_list, driver_error_msg)
            if verbose == 2:
                print(
                    f"\033[1mdriver {driver_name} has {len(driver_paths)} diffrent paths\033[0m"
                )
            numeric_state_error_msg = (
                "please give the correct driving state number, only 3 states available."
            )
            state = self.numeric_to_string_from_list(
                state, self.STATES, error_msg=numeric_state_error_msg
            )
            state_error_msg = f"\nplease give driving state number or give the correct driving state from:\n{self.STATES}"
            self.state = pattern_search(
                state.upper(), self.STATES, state_error_msg
            ).pop()
            state_paths = pattern_search(self.state, driver_paths, state_error_msg)
            numeric_road_error_msg = (
                "please give the correct road number, only 2 roads available."
            )
            road = self.numeric_to_string_from_list(
                road, self.ROADS, numeric_road_error_msg
            )
            road_error_msg = f"\nplease give road number or give the correct road name from:\n{self.ROADS}"
            self.road = pattern_search(road.upper(), self.ROADS, road_error_msg).pop()
            road_paths = pattern_search(self.road, state_paths, road_error_msg)
            road_paths = list(road_paths)
            if len(road_paths) > 1 and verbose == 2:
                print(
                    f"\nThere is more than one path for this case,the selection depended on the standard parameter"
                )
            if standard:
                try:
                    self.file_path = road_paths[0]
                except IndexError:
                    raise ValueError(
                        "driver 6 does not have data for state: AGGRESSIVE and road: SECONDARY"
                    )
            elif not standard:
                try:
                    self.file_path = road_paths[1]
                except IndexError:
                    raise ValueError("Only standard file available for this case.")
            if verbose == 2:
                print(f"\nThe file path is:\n\033[1m'{self.file_path}'\033[0m\n")

    def __call__(self):
        """A method for the class instance call

        Returns:
            * str: The file path.
        """
        return self.file_path

    def numeric_to_string_from_list(self, element, element_list, error_msg):
        """A method for checking if an element is an integer or numeric to return a string from a list matching it's index.

        Args:
            * element (int or str): An element to be checked if it is numeric.
            * element_list (set or list): A list to have matching strings for the numeric element.
            * error_msg (str): An error message to be displayed when a ValueError is raised.

        Raises:
            * ValueError: if the numeric element is not in the range of the element_list.

        Returns:
            * str: The string matching the index of the numeric elemnt or the entry element if the element is not a numeric value.
        """
        if isinstance(element, int) or element.isnumeric():
            if 0 < int(element) < len(element_list) + 1:
                element = element_list[int(element) - 1]
                return element
            else:
                raise ValueError(error_msg)
        return element


class UahDatasetLoader(UahDatasetPathFinder):
    """A class for loading a UahDataset file into a dataframe, includes a dictionary attribute of the `files_column_names` and a `data` attribute, a `__call__ ` method for calling an instance of the class to return the `data` attribute.

    Args:
        * UahDatasetPathFinder (object): A class object for finding the UahDataset path.
        * parent_path (str): The path to the parent directory of the UahDataset.
        * driver_id (str or int): The number of the driver, can be an integer from `1` to `6` or a string that ends with integer in the same range such as `"driver1"` or `"D6"`.
        * state_id (str or int): The state of the driving session, can be an integer from `1` to `3` or a string such as `"normal"`,`"aggressive"` or `"drowsy"` not case sensitive.
        * road_id (str or int): The type of road in the driving session, can be an integer from `1` to `2` or a string such as `"motorway"` or `"secondary"` not case sensitive.
        * data_type_id (str or int): The type of data or the file of data chosen, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
        * standard (bool, optional): if `True` gives you the first session and if `False` gives the second session, of course for both cases it depends on the availability of the data. Defaults to `True`.
        * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``0``.
    """

    files_column_names = {
        "RAW_GPS.txt": [
            "Timestamp (seconds)",
            "Speed (km/h)",
            "Latitude coordinate (degrees)",
            "Longitude coordinate (degrees)",
            "Altitude (meters)",
            "Vertical accuracy (degrees)",
            "Horizontal accuracy (degrees)",
            "Course (degrees)",
            "Difcourse: course variation (degrees)",
        ],
        "RAW_ACCELEROMETERS.txt": [
            "Timestamp (seconds)",
            "Boolean of system activated (1 if >50km/h)",
            "Acceleration in X (Gs)",
            "Acceleration in Y (Gs)",
            "Acceleration in Z (Gs)",
            "Acceleration in X filtered by KF (Gs)",
            "Acceleration in Y filtered by KF (Gs)",
            "Acceleration in Z filtered by KF (Gs)",
            "Roll (degrees)",
            "Pitch (degrees)",
            "Yaw (degrees)",
        ],
        "PROC_LANE_DETECTION.txt": [
            "Timestamp (seconds)",
            "X: car position relative to lane center (meters)",
            "Phi: car angle relative to lane curvature (degrees)",
            "W: road width (meters)",
            "State of the lane det. algorithm [-1=calibrating,0=initializing, 1=undetected, 2=detected/running]",
        ],
        "PROC_VEHICLE_DETECTION.txt": [
            "Timestamp (seconds)",
            "Distance to ahead vehicle in current lane (meters) [value -1 means no car is detected in front]",
            "Time of impact to ahead vehicle (seconds) [distance related to own speed]",
            "Number of detected vehicles in this frame (traffic)",
            "GPS speed (km/h) [same as in RAW GPS]",
        ],
        "PROC_OPENSTREETMAP_DATA.txt": [
            "Timestamp (seconds)",
            "Maximum allowed speed of current road (km/h)",
            "Reliability of obtained maxspeed (0=unknown,1=reliable, 2=used previously obtained maxspeed,3=estimated by type of road)",
            "Type of road (motorway, trunk, secondary...)",
            "Number of lanes in current road",
            "Estimated current lane (1=right lane, 2=first left lane, 3=second left lane, etc) [experimental]",
            "GPS Latitude used to query OSM (degrees)",
            "GPS Longitude used to query OSM (degrees)",
            "OSM delay to answer query (seconds)",
            "GPS speed (km/h) [same as in RAW GPS]",
        ],
        "EVENTS_LIST_LANE_CHANGES.txt": [
            "Timestamp (seconds)",
            "Type [+ indicates right and - left, 1 indicates normal lane change and 2 slow lane change]",
            "GPS Latitude of the event (degrees)",
            "GPS Longitude of the event (degrees)",
            "Duration of the lane change (seconds) [measured since the car position is near the lane marks]",
            "Time threshold to consider irregular change (secs.) [slow if change duration is over this threshold and fast if duration is lower than threshold/3]",
        ],
        "EVENTS_INERTIAL.txt": [
            "Timestamp (seconds)",
            "Type (1=braking, 2=turning, 3=acceleration)",
            "Level (1=low, 2=medium, 3=high)",
            "GPS Latitude of the event",
            "GPS Longitude of the event ",
            "Date of the event in YYYYMMDDhhmmss format",
        ],
        "SEMANTIC_FINAL.txt": [
            "Hour of route start",
            "Minute of route start",
            "Second of route start",
            "Average speed during trip (km/h)",
            "Maximum achieved speed during route (km/h)",
            "Lanex score (internal value, related to lane drifting)",
            "Driving time (in minutes)",
            "Hour of route end",
            "Minute of route end",
            "Second of route end",
            "Trip distance (km)",
            "ScoreLongDist	(internal value, Score accelerations)",
            "ScoreTranDist	(internal value, Score turnings)",
            "ScoreSpeedDist (internal value, Score brakings)",
            "ScoreGlobal (internal value, old score that is not used anymore)",
            "Alerts Long (internal value)",
            "Alerts Late (internal value)",
            "Alerts Lanex (internal value)",
            "Number of vehicle stops during route (experimental, related to fuel efficiency estimation)",
            "Speed variability (experimental, related to fuel efficiency estimation)",
            "Acceleration noise (experimental, related to fuel efficiency estimation)",
            "Kinetic energy (experimental, related to fuel efficiency estimation)",
            "Driving time (in seconds)",
            "Number of curves in the route",
            "Power exherted (experimental, related to fuel efficiency estimation)",
            "Acceleration events (internal value)",
            "Braking events (internal value)",
            "Turning events (internal value)",
            "Longitudinal-distraction Global Score (internal value, combines mean[31] and std[32])",
            "Transversal-distraction Global Score (internal value, combines mean[33] and std[34])",
            "Mean Long.-dist. score (internal value)",
            "STD Long.-dist. score (internal value)",
            "Average Trans.-dist. score (internal value)",
            "STD Trans.-dist. score (internal value)",
            "Lacc (number of low accelerations)",
            "Macc (number of medium accelerations)",
            "Hacc (number of high accelerations)",
            "Lbra (number of low brakings)",
            "Mbra (number of medium brakings)",
            "Hbra (number of high brakings)",
            "Ltur (number of low turnings)",
            "Mtur (number of medium turnings)",
            "Htur (number of high turnings)",
            "Score total (base 100, direct mean of the other 7 scores [45-51])",
            "Score accelerations (base 100)",
            "Score brakings (base 100)",
            "Score turnings (base 100)",
            "Score lane-weaving (base 100)",
            "Score lane-drifting (base 100)",
            "Score overspeeding (base 100)",
            "Score car-following (base 100)",
            "Ratio normal (base 1)",
            "Ratio drowsy (base 1)",
            "Ratio aggressive (base 1)",
        ],
        "SEMANTIC_ONLINE.txt": [
            "TimeStamp since route start (seconds)",
            "GPS Latitude (degrees)",
            "GPS Longitude (degrees)",
            "Score total WINDOW (base 100, direct mean of the other 7 scores)",
            "Score accelerations WINDOW (base 100)",
            "Score brakings WINDOW (base 100)",
            "Score turnings WINDOW (base 100)",
            "Score weaving WINDOW (base 100)",
            "Score drifting WINDOW (base 100)",
            "Score overspeeding WINDOW (base 100)",
            "Score car-following WINDOW (base 100)",
            "Ratio normal WINDOW (base 1)",
            "Ratio drowsy WINDOW (base 1)",
            "Ratio aggressive WINDOW (base 1)",
            "Ratio distracted WINDOW (1=distraction detected in last 2 seconds, 0=otherwise)",
            "Score total (base 100, direct mean of the other 7 scores)",
            "Score accelerations (base 100)",
            "Score brakings (base 100)",
            "Score turnings (base 100)",
            "Score weaving (base 100)",
            "Score drifting (base 100)",
            "Score overspeeding (base 100)",
            "Score car-following (base 100)",
            "Ratio normal (base 1)",
            "Ratio drowsy (base 1)",
            "Ratio aggressive (base 1)",
            "Ratio distracted (1=distraction detected in last 2 seconds, 0=otherwise)",
        ],
    }

    def __init__(
        self,
        parent_path,
        driver_id,
        state_id,
        road_id,
        data_type_id,
        standard=True,
        verbose=0,
    ):
        """A constructor for the UahDatasetLoader class.

        Args:
        * parent_path (str): The path to the parent directory of the UahDataset.
        * driver_id (str or int): The number of the driver, can be an integer from `1` to `6` or a string that ends with integer in the same range such as `"driver1"` or `"D6"`.
        * state_id (str or int): The state of the driving session, can be an integer from `1` to `3` or a string such as `"normal"`,`"aggressive"` or `"drowsy"` not case sensitive.
        * road_id (str or int): The type of road in the driving session, can be an integer from `1` to `2` or a string such as `"motorway"` or `"secondary"` not case sensitive.
        * data_type_id (str or int): The type of data or the file of data chosen, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
        * standard (bool, optional): if `True` gives you the first session and if `False` gives the second session, of course for both cases it depends on the availability of the data. Defaults to `True`.
        * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1``. Defaults to ``0``.
        """
        super().__init__(
            parent_path,
            driver_id,
            state_id,
            road_id,
            data_type_id,
            standard,
            verbose=verbose,
        )
        if self.file == "SEMANTIC_FINAL.txt":
            delimiter = None
        else:
            delimiter = " "
        self.data = read_csv(self.file_path, delimiter=delimiter, header=None)
        if verbose == 1:
            print(f"\033[1m+++Data loaded successfully from {self.file} +++\033[0m\n")
        if self.file == self.SENSOR_FILES[0]:
            self.data = self.data.drop([9, 10, 11, 12], axis=1)
        elif self.file == self.SENSOR_FILES[1]:
            try:
                self.data = self.data.drop(11, axis=1)
            except Exception:
                self.data = self.data
        elif self.file == self.SENSOR_FILES[3]:
            self.data = self.data.drop(5, axis=1)
        elif self.file == self.SENSOR_FILES[4]:
            self.data = self.data.drop(10, axis=1)
        elif self.file == self.SENSOR_FILES[6]:
            self.data = self.data.drop(6, axis=1)
        elif self.file == self.SENSOR_FILES[7]:
            self.data = self.data.transpose()
        elif self.file == self.SENSOR_FILES[8]:
            self.data = self.data.drop(27, axis=1)
        self.data.columns = self.files_column_names[self.file]

    def __call__(self):
        """A method for the class instance call

        Returns:
            * pandas.DataFrame: The loaded data dataframe.
        """
        return self.data


class UAHDatasetBuilder:
    """A class to build a time series dataset from the UAHDataset using two files.includes a `data` attribute for the built UAHDataset, a `__call__ ` method for calling an instance of the class to return the `data` attribute.
    Args:
        * path (str): The path to the parent directory of the UahDataset.
        * datatype1 (str or int): The type of data or the file of data chosen for the first file, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
        * datatype2 (str or int): The type of data or the file of data chosen for the first file, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
        * window_size_dt1 (int, optional): The size of the first data window in the case of window stepping the data, if set to `0` will not perform the window stepping and if set to another value will perform window steping with mean transformation on the data. Defaults to `0`.
        * step_dt1 (int, optional): The length of the step for window stepping the first data, if smaller than `window_size_dt1` will result in overlapping windows, if equal to `window_size_dt1` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `0`.
        * window_size_dt2 (int, optional): The size of the second data window in the case of window stepping the data, if set to `0` will not perform the window stepping and if set to another value will perform window steping with mean transformation on the data. Defaults to `0`.
        * step_dt2 (int, optional): The length of the step for window stepping the second data, if smaller than `window_size_dt2` will result in overlapping windows, if equal to `window_size_dt2` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `0`.
        * save_to (str, optional): Can be `"csv"` or `"parquet"` or `None` if set to `"csv"` or `"parquet"` will save the dataset to the corresponding format in a newly created directory, if set to `None` will not save the dataset. Defaults to `None`.
        * name_dataset (str, optional): The name of the dataset in case of save_to is set to save the dataset. Defaults to "UAHDataset".
        * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1`` or ``2``or ``3``. Defaults to ``0``.
    """

    def __init__(
        self,
        path,
        datatype1,
        datatype2,
        window_size_dt1=0,
        step_dt1=0,
        window_size_dt2=0,
        step_dt2=0,
        save_to=None,
        name_dataset="UAHDataset",
        verbose=0,
    ):
        """A constructor for the UAHDatasetBuilder class.

        Args:
            * path (str): The path to the parent directory of the UahDataset.
            * datatype1 (str or int): The type of data or the file of data chosen for the first file, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
            * datatype2 (str or int): The type of data or the file of data chosen for the first file, can be an integer from 1 to 9 or a string such as `"gps"` or `"accelerometer"` not case sensitive.
            * window_size_dt1 (int, optional): The size of the first data window in the case of window stepping the data, if set to `0` will not perform the window stepping and if set to another value will perform window steping with mean transformation on the data. Defaults to `0`.
            * step_dt1 (int, optional): The length of the step for window stepping the first data, if smaller than `window_size_dt1` will result in overlapping windows, if equal to `window_size_dt1` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `0`.
            * window_size_dt2 (int, optional): The size of the second data window in the case of window stepping the data, if set to `0` will not perform the window stepping and if set to another value will perform window steping with mean transformation on the data. Defaults to `0`.
            * step_dt2 (int, optional): The length of the step for window stepping the second data, if smaller than `window_size_dt2` will result in overlapping windows, if equal to `window_size_dt2` performs standard window stepping, if bigger will skip some rows (not recommended). Defaults to `0`.
            * save_to (str, optional): Can be `"csv"` or `"parquet"` or `None` if set to `"csv"` or `"parquet"` will save the dataset to the corresponding format in a newly created directory, if set to `None` will not save the dataset. Defaults to `None`.
            * name_dataset (str, optional): The name of the dataset in case of save_to is set to save the dataset. Defaults to "UAHDataset".
            * verbose (int, optional): An integer of the verbosity of the operation can be ``0`` or ``1`` or ``2``or ``3``. Defaults to ``0``.
        """
        self.data = pd.DataFrame()
        count = 0
        for i in (1, 2, 3, 4, 5, 6):
            for j in (1, 2, 3):
                for k in (1, 2):
                    for standard in True, False:
                        try:
                            if verbose > 0:
                                print(
                                    f"\n\033[1m<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< CASE{count+1} STARTS >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\033[0m\n"
                                )
                                print(
                                    "\n===========LOADING THE FIRST DATA FILE===========\n"
                                )
                            raw_data1 = UahDatasetLoader(
                                path,
                                i,
                                j,
                                k,
                                datatype1,
                                standard=standard,
                                verbose=verbose,
                            )
                            if verbose > 0:
                                print(
                                    "\n===========CLEANING THE FIRST FILE DATA ===========\n"
                                )
                            data_chunk1 = DataCleaner(
                                raw_data1(),
                                window_size=window_size_dt1,
                                step=step_dt1,
                                window_transformation=True,
                                verbose=0,
                            ).dataset
                            if verbose > 0:
                                print(
                                    f"\033[1mFirst data shape is: {data_chunk1.shape}\033[0m"
                                )
                                print(
                                    "\n===========LOADING THE SECOND DATA FILE===========\n"
                                )
                            raw_data2 = UahDatasetLoader(
                                path,
                                i,
                                j,
                                k,
                                datatype2,
                                standard=standard,
                                verbose=verbose,
                            )
                            if verbose > 0:
                                print(
                                    "\n===========CLEANING THE SECOND FILE DATA ===========\n"
                                )
                            data_chunk2 = DataCleaner(
                                raw_data2(),
                                window_size=window_size_dt2,
                                step=step_dt2,
                                window_transformation=True,
                                verbose=0,
                            ).dataset
                            if verbose > 0:
                                print(
                                    f"\033[1mSecond data shape is: {data_chunk2.shape}\033[0m"
                                )
                                print("\n===========MERGING DATA===========\n")
                            merge_verbose = 1 if verbose > 2 else 0
                            data_chunk_merged = DataCleaner(
                                data=data_chunk1,
                                merge_data=data_chunk2,
                                add_columns_dictionnary={
                                    "driver": raw_data1.driver,
                                    "road": raw_data1.road,
                                    "target": raw_data1.state,
                                },
                                verbose=merge_verbose,
                            ).dataset
                            if verbose > 0:
                                print(
                                    f"\033[1mMerged data shape is: {data_chunk_merged.shape}\033[0m"
                                )
                            self.data = self.data.append(data_chunk_merged)
                            self.data = self.data.reset_index()
                            self.data = self.data.drop("index", axis=1)
                            count += 1
                            if verbose > 0:
                                print(
                                    f"\n\033[1m<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< CASE{count} FINISHED >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\033[0m\n"
                                )
                        except ValueError:
                            if verbose > 0:
                                print(
                                    "\033[1mNO EXISTING FILE WITH THIS PATH...CONTINUE CASE WITH ANOTHER PATH\033[0m\n"
                                )
                                print(
                                    f"\n\033[1m<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< NEW CASE{count+1} ATTEMPT >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\033[0m\n"
                                )
        if verbose > 0:
            print(f"\n\033[1m...DATA BUILDING COMPLETE WITH {count} CASES\033[0m\n")
            print(f"\033[1mFINAL DATA SHAPE: {self.data.shape}\033[0m")
        PATH = "dataset"
        if save_to == "csv":
            save_csv(self.data, PATH, name_dataset, verbose=verbose)
        elif save_to == "parquet":
            save_parquet(self.data, PATH, name_dataset, verbose=verbose)

    def __call__(self):
        """A method for the class instance call

        Returns:
            * pandas.DataFrame: built dataset dataframe.
        """
        return self.data


if __name__ == "__main__":
    DATA_DIR_PATH = "/run/media/najem/34b207a8-0f0c-4398-bba2-f31339727706/home/stock/The_stock/dev & datasets/PhD/datasets/UAH-DRIVESET-v1/"
    dataset = UAHDatasetBuilder(
        DATA_DIR_PATH,
        "gps",
        "acc",
        window_size_dt2=10,
        step_dt2=10,
        verbose=2,
    )

# %%
